import os

import requests


def format_author_name(nm):
    """
    make sure all words in name start with a capital
    """
    nms = nm.split(" ")
    auth_nm = " ".join([n.capitalize() for n in nms])
    return auth_nm


def get_github_profile_name(email):
    """
    given an email addr return the github login name
    """
    email = email.replace(" ", "")
    nm = email.split("@")[0]
    return nm


def get_file_contributors_via_api(repo_name, file_path):
    """
    This function takes the repo name (ex:"definitions") and relative path to the nxdl
    file (ex: "applications/NXmx.nxdl.xml") and using the github api it retrieves a dictionary
    of committers for that file in descending date order.

    In order to increase the capacity (rate) of use of the github API an access token is used if it exists
    as an environment variable called GH_TOKEN, in the ci yaml file this is expected to be assigned from the secret
    object like this
    env:
          GH_TOKEN: ${{ secrets.GH_TOKEN }}

    With the access token the rate is 5000 times per hour and without it is 60

    returns a sorted dict of unique contributors to a file, or None if no GH_TOKEN has been defined in os.environ
    """
    have_token = False
    access_token = os.getenv("GH_TOKEN")
    if (
        access_token is not None and access_token != "NONE"
    ):  # latter clause is false in most CI cases
        if len(access_token) > 0:
            have_token = True
    else:
        # because the environment does not contain GH_TOKEN, assume the user wants to build the
        # docs without contributor info
        return None

    contrib_skip_list = ["GitHub"]
    url = f"https://api.github.com/repos/nexusformat/{repo_name}/commits"
    params = {"path": file_path}
    headers = {}
    if have_token:
        # Set the headers with the access token
        headers = {
            "Authorization": f"token {access_token}",
            "Accept": "application/vnd.github.v3+json",
        }

    response = requests.get(url, params=params, headers=headers)
    commits = response.json()
    if response.status_code != 200:
        # if its 403: the max rate per hour has been reached
        raise Exception(
            f"access_token={access_token}, {commits['message']},{commits['documentation_url']}"
        )

    contributor_names = set()
    contribs_dct = {}
    _email_lst = []
    for commit_dct in commits:
        if commit_dct["committer"] is not None:
            contributor = commit_dct["commit"]["committer"]["name"]
            if contributor in contrib_skip_list:
                continue
            contributor_names.add(contributor)
            if commit_dct["commit"]["committer"]["email"] not in _email_lst:
                _email = commit_dct["commit"]["committer"]["email"]
                _email_lst.append(_email)
                contribs_dct[commit_dct["commit"]["committer"]["date"]] = {
                    "name": format_author_name(
                        commit_dct["commit"]["committer"]["name"]
                    ),
                    "commit_dct": commit_dct,
                }

    # sort them so they are in descending order from newest to oldest
    sorted_keys = sorted(contribs_dct.keys(), reverse=True)
    sorted_dict = {key: contribs_dct[key] for key in sorted_keys}

    return sorted_dict



from pathlib import Path
import h5py

filename = str(
    Path(__file__).absolute().parent.parent
    / "simple_example_basic"
    / "simple_example_basic.nexus.hdf5"
)
with h5py.File(filename, "r") as nx:
    # find the default NXentry group
    nx_entry = nx[nx.attrs["default"]]
    # find the default NXdata group
    nx_data = nx_entry[nx_entry.attrs["default"]]
    # find the signal field
    signal = nx_data[nx_data.attrs["signal"]]
    # find the axes field(s)
    attr_axes = nx_data.attrs["axes"]
    if isinstance(attr_axes, (set, tuple, list)):
        #  but check that attr_axes only describes 1-D data
        if len(attr_axes) == 1:
            attr_axes = attr_axes[0]
        else:
            raise ValueError(f"expected 1-D data but @axes={attr_axes}")
    axes = nx_data[attr_axes]

    print(f"file: {nx.filename}")
    print(f"signal: {signal.name}")
    print(f"axes: {axes.name}")
    print(f"{axes.name} {signal.name}")
    for x, y in zip(axes, signal):
        print(x, y)



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Definitions for different conventions associated with Euler angles."""

# pylint: disable=no-member

# Euler angles are a parameterization for orientations which use three
# consecutively rotations (3D) rotations to rotate a configuration
# described by an attached Cartesian CS na(False, "", ""),         "", ""),med A into configuration
# described by an attached Cartesian CS named B
# the sequence about which specific axis and new/intermediate axes
# one rotates enables to distinguish different types of Euler angle
# so-called Euler-angle conventions, most commonly used in materials
# science is the convention of H.-J. Bunge aka zxz convention

# "Bunge" https://doi.org/10.1016/C2013-0-11769-2
# "Rowenhorst" https://doi.org/10.1088/0965-0393/23/8/083501
# "Morawiec" https://doi.org/10.1007/978-3-662-09156-2
# "Britton" https://doi.org/10.1016/j.matchar.2016.04.008


which_euler_convention = {"xxx": (False, "", ""),
                          "xxy": (True, "", ""),
                          "xxz": (True, "", ""),
                          "xyx": (True, "", ""),
                          "xyy": (True, "", ""),
                          "xyz": (True, "", ""),
                          "xzx": (True, "", ""),
                          "xzy": (True, "", ""),
                          "xzz": (True, "", ""),
                          "yxx": (True, "", ""),
                          "yxy": (True, "", ""),
                          "yxz": (True, "", ""),
                          "yyx": (True, "", ""),
                          "yyy": (False, "", ""),
                          "yyz": (True, "", ""),
                          "yzx": (True, "", ""),
                          "yzy": (True, "", ""),
                          "yzz": (True, "", ""),
                          "zxx": (True, "", ""),
                          "zxy": (True, "", ""),
                          "zxz": (True, "Bunge", "proper"),
                          "zyx": (True, "", ""),
                          "zyy": (True, "", ""),
                          "zyz": (True, "", ""),
                          "zzx": (True, "", ""),
                          "zzy": (True, "", ""),
                          "zzz": (False, "", "")}



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Test cases for the helper functions used by the DataConverter."""

import xml.etree.ElementTree as ET
import os
import logging
from setuptools import distutils
import pytest
import numpy as np

from pynxtools.dataconverter import helpers
from pynxtools.dataconverter.template import Template


def remove_optional_parent(data_dict: Template):
    """Completely removes the optional group from the test Template."""
    internal_dict = Template(data_dict)
    del internal_dict["/ENTRY[my_entry]/optional_parent/required_child"]
    del internal_dict["/ENTRY[my_entry]/optional_parent/optional_child"]
    del internal_dict["/ENTRY[my_entry]/optional_parent/req_group_in_opt_group/DATA[data]"]

    return internal_dict


def alter_dict(data_dict: Template, key: str, value: object):
    """Helper function to alter a single entry in dict for parametrize."""
    if data_dict is not None:
        internal_dict = Template(data_dict)
        internal_dict[key] = value
        return internal_dict

    return None


def set_to_none_in_dict(data_dict: Template, key: str, optionality: str):
    """Helper function to forcefully set path to 'None'"""
    if data_dict is not None:
        internal_dict = Template(data_dict)
        internal_dict[optionality][key] = None
        return internal_dict

    return None


def remove_from_dict(data_dict: Template, key: str, optionality: str = 'optional'):
    """Helper function to remove a key from dict"""
    if data_dict is not None and key in data_dict[optionality]:
        internal_dict = Template(data_dict)
        del internal_dict[optionality][key]
        return internal_dict

    return None


def listify_template(data_dict: Template):
    """Helper function to turn most values in the Template into lists"""
    listified_template = Template()
    for optionality in ("optional", "recommended", "required", "undocumented"):
        for path in data_dict[optionality]:
            if path[path.rindex("/") + 1:] in ("@units", "type", "definition", "date_value"):
                listified_template[optionality][path] = data_dict[optionality][path]
            else:
                listified_template[optionality][path] = [data_dict[optionality][path]]
    return listified_template


@pytest.fixture(name="nxdl_root")
def fixture_nxdl_root():
    """pytest fixture to load the same NXDL file for all tests."""
    nxdl_file = os.path.join("tests", "data", "dataconverter", "NXtest.nxdl.xml")
    yield ET.parse(nxdl_file).getroot()


@pytest.fixture(name="template")
def fixture_template():
    """pytest fixture to use the same template in all tests"""
    nxdl_root = ET.parse("tests/data/dataconverter/NXtest.nxdl.xml").getroot()
    template = Template()
    helpers.generate_template_from_nxdl(nxdl_root, template)
    yield template


@pytest.mark.usefixtures("template")
@pytest.fixture(name="filled_test_data")
def fixture_filled_test_data(template, tmp_path):
    """pytest fixture to setup a filled in template."""

    # Copy original measurement file to tmp dir,
    # because h5py.ExternalLink is modifying it while
    # linking the nxs file.
    distutils.file_util.copy_file(f"{os.path.dirname(__file__)}"
                                  f"/../"
                                  f"data/dataconverter/"
                                  f"readers/mpes/"
                                  f"xarray_saved_small_calibration.h5",
                                  tmp_path)

    template.clear()
    template["/ENTRY[my_entry]/NXODD_name/float_value"] = 2.0
    template["/ENTRY[my_entry]/NXODD_name/float_value/@units"] = "nm"
    template["/ENTRY[my_entry]/optional_parent/required_child"] = 1
    template["/ENTRY[my_entry]/optional_parent/optional_child"] = 1
    template["/ENTRY[my_entry]/NXODD_name/bool_value"] = True
    template["/ENTRY[my_entry]/NXODD_name/int_value"] = 2
    template["/ENTRY[my_entry]/NXODD_name/int_value/@units"] = "eV"
    template["/ENTRY[my_entry]/NXODD_name/posint_value"] = np.array([1, 2, 3],
                                                                    dtype=np.int8)
    template["/ENTRY[my_entry]/NXODD_name/posint_value/@units"] = "kg"
    template["/ENTRY[my_entry]/NXODD_name/char_value"] = "just chars"
    template["/ENTRY[my_entry]/definition"] = "NXtest"
    template["/ENTRY[my_entry]/definition/@version"] = "2.4.6"
    template["/ENTRY[my_entry]/program_name"] = "Testing program"
    template["/ENTRY[my_entry]/NXODD_name/type"] = "2nd type"
    template["/ENTRY[my_entry]/NXODD_name/date_value"] = ("2022-01-22T12"
                                                          ":14:12.05018+00:00")
    template["/ENTRY[my_entry]/required_group/description"] = "An example description"
    template["/ENTRY[my_entry]/required_group2/description"] = "An example description"
    template["/ENTRY[my_entry]/does/not/exist"] = "random"
    template["/ENTRY[my_entry]/links/ext_link"] = {"link":
                                                   f"{tmp_path}/"
                                                   f"xarray_saved_small_cali"
                                                   f"bration.h5:/axes/ax3"
                                                   }
    yield template


TEMPLATE = Template()
TEMPLATE["optional"]["/ENTRY[my_entry]/NXODD_name/float_value"] = 2.0  # pylint: disable=E1126
TEMPLATE["optional"]["/ENTRY[my_entry]/NXODD_name/float_value/@units"] = "nm"  # pylint: disable=E1126
TEMPLATE["optional"]["/ENTRY[my_entry]/optional_parent/required_child"] = 1  # pylint: disable=E1126
TEMPLATE["optional"]["/ENTRY[my_entry]/optional_parent/optional_child"] = 1  # pylint: disable=E1126
TEMPLATE["required"]["/ENTRY[my_entry]/NXODD_name/bool_value"] = True  # pylint: disable=E1126
TEMPLATE["required"]["/ENTRY[my_entry]/NXODD_name/int_value"] = 2  # pylint: disable=E1126
TEMPLATE["required"]["/ENTRY[my_entry]/NXODD_name/int_value/@units"] = "eV"  # pylint: disable=E1126
TEMPLATE["required"]["/ENTRY[my_entry]/NXODD_name/posint_value"] = np.array([1, 2, 3],  # pylint: disable=E1126
                                                                            dtype=np.int8)  # pylint: disable=E1126
TEMPLATE["required"]["/ENTRY[my_entry]/NXODD_name/posint_value/@units"] = "kg"  # pylint: disable=E1126
TEMPLATE["required"]["/ENTRY[my_entry]/NXODD_name/char_value"] = "just chars"  # pylint: disable=E1126
TEMPLATE["required"]["/ENTRY[my_entry]/definition"] = "NXtest"  # pylint: disable=E1126
TEMPLATE["required"]["/ENTRY[my_entry]/definition/@version"] = "2.4.6"  # pylint: disable=E1126
TEMPLATE["required"]["/ENTRY[my_entry]/program_name"] = "Testing program"  # pylint: disable=E1126
TEMPLATE["required"]["/ENTRY[my_entry]/NXODD_name/type"] = "2nd type"  # pylint: disable=E1126
TEMPLATE["required"]["/ENTRY[my_entry]/NXODD_name/date_value"] = "2022-01-22T12:14:12.05018+00:00"  # pylint: disable=E1126
TEMPLATE["optional"]["/ENTRY[my_entry]/required_group/description"] = "An example description"
TEMPLATE["optional"]["/ENTRY[my_entry]/required_group2/description"] = "An example description"
TEMPLATE["required"]["/ENTRY[my_entry]/optional_parent/req_group_in_opt_group/DATA[data]"] = 1
TEMPLATE["lone_groups"] = ['/ENTRY[entry]/required_group',
                           '/ENTRY[entry]/required_group2',
                           '/ENTRY[entry]/optional_parent/req_group_in_opt_group']
TEMPLATE["optional"]["/@default"] = "Some NXroot attribute"


@pytest.mark.parametrize("data_dict,error_message", [
    pytest.param(
        alter_dict(TEMPLATE, "/ENTRY[my_entry]/NXODD_name/int_value", "not_a_num"),
        ("The value at /ENTRY[my_entry]/NXODD_name/in"
         "t_value should be of Python type: (<class 'int'>, <cla"
         "ss 'numpy.ndarray'>, <class 'numpy.signedinteger'>),"
         " as defined in the NXDL as NX_INT."),
        id="string-instead-of-int"),
    pytest.param(
        alter_dict(TEMPLATE, "/ENTRY[my_entry]/NXODD_name/bool_value", "NOT_TRUE_OR_FALSE"),
        ("The value at /ENTRY[my_entry]/NXODD_name/bool_value sh"
         "ould be of Python type: (<class 'bool'>, <class 'numpy.ndarray'>, <class '"
         "numpy.bool_'>), as defined in the NXDL as NX_BOOLEAN."),
        id="string-instead-of-int"),
    pytest.param(
        alter_dict(TEMPLATE, "/ENTRY[my_entry]/NXODD_name/int_value", {"link": "/a-link"}),
        (""),
        id="link-dict-instead-of-bool"),
    pytest.param(
        alter_dict(TEMPLATE, "/ENTRY[my_entry]/NXODD_name/posint_value", -1),
        ("The value at /ENTRY[my_entry]/NXODD_name/posint_value "
         "should be a positive int."),
        id="negative-posint"),
    pytest.param(
        alter_dict(TEMPLATE, "/ENTRY[my_entry]/NXODD_name/char_value", 3),
        ("The value at /ENTRY[my_entry]/NXODD_name/char_value should be of Python type:"
         " (<class 'str'>, <class 'numpy.ndarray'>, <class 'numpy.chararray'>),"
         " as defined in the NXDL as NX_CHAR."),
        id="int-instead-of-chars"),
    pytest.param(
        alter_dict(TEMPLATE, "/ENTRY[my_entry]/NXODD_name/float_value", None),
        "",
        id="empty-optional-field"),
    pytest.param(
        set_to_none_in_dict(TEMPLATE, "/ENTRY[my_entry]/NXODD_name/bool_value", "required"),
        ("The data entry corresponding to /ENTRY[entry]/NXODD_name/bool_value is"
         " required and hasn't been supplied by the reader."),
        id="empty-required-field"),
    pytest.param(
        alter_dict(TEMPLATE,
                   "/ENTRY[my_entry]/NXODD_name/date_value",
                   "2022-01-22T12:14:12.05018+00:00"),
        "",
        id="UTC-with-+00:00"),
    pytest.param(
        alter_dict(TEMPLATE,
                   "/ENTRY[my_entry]/NXODD_name/date_value",
                   "2022-01-22T12:14:12.05018Z"),
        "",
        id="UTC-with-Z"),
    pytest.param(
        alter_dict(TEMPLATE,
                   "/ENTRY[my_entry]/NXODD_name/date_value",
                   "2022-01-22T12:14:12.05018-00:00"),
        "The date at /ENTRY[my_entry]/NXODD_name/date_value should be a timezone aware"
        " ISO8601 formatted str. For example, 2022-01-22T12:14:12.05018Z or 2022-01-22"
        "T12:14:12.05018+00:00.",
        id="UTC-with--00:00"),
    pytest.param(
        listify_template(TEMPLATE),
        "",
        id="lists"),
    pytest.param(
        alter_dict(TEMPLATE, "/ENTRY[my_entry]/NXODD_name/type", "Wrong option"),
        ("The value at /ENTRY[my_entry]/NXODD_name/type should be on of the following"
         " strings: [1st type,2nd type,3rd type,4th type]"),
        id="wrong-enum-choice"),
    pytest.param(
        set_to_none_in_dict(TEMPLATE,
                            "/ENTRY[my_entry]/optional_parent/required_child",
                            "optional"),
        ("The data entry, /ENTRY[my_entry]/optional_parent/optional_child, has an "
         "optional parent, /ENTRY[entry]/optional_parent, with required children set"
         ". Either provide no children for /ENTRY[entry]/optional_parent or provide "
         "all required ones."),
        id="atleast-one-required-child-not-provided-optional-parent"),
    pytest.param(
        alter_dict(alter_dict(TEMPLATE,
                              "/ENTRY[my_entry]/optional_parent/required_child",
                              None),
                   "/ENTRY[my_entry]/optional_parent/optional_child",
                   None),
        (""),
        id="no-child-provided-optional-parent"),
    pytest.param(
        TEMPLATE,
        "",
        id="valid-data-dict"),
    pytest.param(
        remove_from_dict(TEMPLATE, "/ENTRY[my_entry]/required_group/description"),
        "The required group, /ENTRY[entry]/required_group, hasn't been supplied.",
        id="missing-empty-yet-required-group"),
    pytest.param(
        remove_from_dict(TEMPLATE, "/ENTRY[my_entry]/required_group2/description"),
        "The required group, /ENTRY[entry]/required_group2, hasn't been supplied.",
        id="missing-empty-yet-required-group2"),
    pytest.param(
        alter_dict(
            remove_from_dict(TEMPLATE, "/ENTRY[my_entry]/required_group/description"),
            "/ENTRY[my_entry]/required_group",
            {}
        ),
        (""),
        id="allow-required-and-empty-group"
    ),
    pytest.param(
        remove_from_dict(TEMPLATE,
                         "/ENTRY[my_entry]/optional_parent/req_group_in_opt_group/DATA[data]",
                         "required"
                         ),
        ("The required group, /ENTRY[entry]/optional_parent/req_group_in_opt_group, hasn't been "
         "supplied while its optional parent, /ENTRY[entry]/optional_parent/"
         "req_group_in_opt_group, is supplied."),
        id="req-group-in-opt-parent-removed"
    ),
    pytest.param(
        remove_optional_parent(TEMPLATE),
        (""),
        id="opt-group-completely-removed"
    ),
])
def test_validate_data_dict(data_dict, error_message, template, nxdl_root, request):
    """Unit test for the data validation routine"""
    if request.node.callspec.id in ("valid-data-dict",
                                    "lists",
                                    "empty-optional-field",
                                    "UTC-with-+00:00",
                                    "UTC-with-Z",
                                    "no-child-provided-optional-parent",
                                    "int-instead-of-chars",
                                    "link-dict-instead-of-bool",
                                    "allow-required-and-empty-group",
                                    "opt-group-completely-removed"):
        helpers.validate_data_dict(template, data_dict, nxdl_root)
    else:
        with pytest.raises(Exception) as execinfo:
            helpers.validate_data_dict(template, data_dict, nxdl_root)
        assert (error_message) == str(execinfo.value)


@pytest.mark.parametrize("nxdl_path,expected", [
    pytest.param(
        "/ENTRY/definition/@version",
        (True, "/ENTRY[entry]/definition/@version"),
        id="path-exists-in-dict"),
    pytest.param(
        "/RANDOM/does/not/@exist",
        (False, None),
        id="path-does-not-exist-in-dict")
])
def test_path_in_data_dict(nxdl_path, expected, template):
    """Unit test for helper function to check if an NXDL path exists in the reader dictionary."""
    assert helpers.path_in_data_dict(nxdl_path, template) == expected


def test_atom_type_extractor_and_hill_conversion():
    """
        Test atom type extractor and conversion to hill
    """

    test_chemical_formula = "(C38H54S4)n(NaO2)5(CH4)NH3B"
    expected_atom_types = ['C', 'H', 'B', 'N', 'Na', 'O', 'S']

    atom_list = helpers.extract_atom_types(test_chemical_formula)

    assert expected_atom_types == atom_list


def test_writing_of_root_attributes(caplog):
    """
    Tests if all root attributes are populated
    """
    template = Template()
    filename = "my_nexus_file.nxs"
    with caplog.at_level(logging.WARNING):
        helpers.add_default_root_attributes(template, filename)

    assert "" == caplog.text

    keys_added = template.keys()
    assert "/@NX_class" in keys_added
    assert template["/@NX_class"] == "NXroot"
    assert "/@file_name" in keys_added
    assert template["/@file_name"] == filename
    assert "/@file_time" in keys_added
    assert "/@file_update_time" in keys_added
    assert "/@NeXus_repository" in keys_added
    assert "/@NeXus_version" in keys_added
    assert "/@HDF5_version" in keys_added
    assert "/@h5py_version" in keys_added


def test_warning_on_root_attribute_overwrite(caplog):
    """
    A warning is emitted when a root attribute is overwritten
    by pynxtools.
    """
    template = Template()
    template["/@NX_class"] = "NXwrong"
    filname = "my_nexus_file.nxs"
    with caplog.at_level(logging.WARNING):
        helpers.add_default_root_attributes(template, filname)
    error_text = (
        "The NXroot entry '/@NX_class' (value: NXwrong) should not be populated by the reader. "
        "This is overwritten by the actually used value 'NXroot'"
    )
    assert error_text in caplog.text

    assert "/@NX_class" in template.keys()
    assert template["/@NX_class"] == "NXroot"



#!/usr/bin/env python3
"""This file collects the function used in the reverse tool nxdl2yaml.

"""
# -*- coding: utf-8 -*-
#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import sys
from typing import List, Dict
import xml.etree.ElementTree as ET
import os

from pynxtools.nyaml2nxdl.nyaml2nxdl_helper import (get_node_parent_info,
                                                    get_yaml_escape_char_dict,
                                                    cleaning_empty_lines)
from pynxtools.dataconverter.helpers import remove_namespace_from_tag


DEPTH_SIZE = "  "
CMNT_TAG = '!--'


def separate_pi_comments(input_file):
    """
    Separate PI comments from ProcessesInstruction (pi)
    """
    comments_list = []
    comment = []
    xml_lines = []

    with open(input_file, "r", encoding='utf-8') as file:
        lines = file.readlines()
        has_pi = True
        for line in lines:
            c_start = '<!--'
            cmnt_end = '-->'
            def_tag = '<definition'

            if c_start in line and has_pi:
                line = line.replace(c_start, '')
                if cmnt_end in line:
                    line = line.replace(cmnt_end, '')
                    comments_list.append(line)
                else:
                    comment.append(line)
            elif cmnt_end in line and len(comment) > 0 and has_pi:
                comment.append(line.replace(cmnt_end, ''))
                comments_list.append(''.join(comment))
                comment = []
            elif def_tag in line or not has_pi:
                has_pi = False
                xml_lines.append(line)
            elif len(comment) > 0 and has_pi:
                comment.append(line)
            else:
                xml_lines.append(line)
    return comments_list, ''.join(xml_lines)


# Collected: https://dustinoprea.com/2019/01/22/python-parsing-xml-and-retaining-the-comments/
class _CommentedTreeBuilder(ET.TreeBuilder):

    def comment(self, text):
        """
        defining comment builder in TreeBuilder
        """
        self.start('!--', {})
        self.data(text)
        self.end('--')


def parse(filepath):
    """
        Construct parse function for modified tree builder for including modified TreeBuilder
        and rebuilding XMLParser.
    """
    comments, xml_str = separate_pi_comments(filepath)
    ctb = _CommentedTreeBuilder()
    xp_parser = ET.XMLParser(target=ctb)
    root = ET.fromstring(xml_str, parser=xp_parser)
    return comments, root


def handle_mapping_char(text, depth=-1, skip_n_line_on_top=False):
    """Check for ":" char and replace it by "':'". """

    escape_char = get_yaml_escape_char_dict()
    for esc_key, val in escape_char.items():
        if esc_key in text:
            text = text.replace(esc_key, val)
    if not skip_n_line_on_top:
        if depth > 0:
            text = add_new_line_with_pipe_on_top(text, depth)
        else:
            raise ValueError("Need depth size to co-ordinate text line in yaml file.")
    return text


def add_new_line_with_pipe_on_top(text, depth):
    """
    Return modified text for what we get error in converter, such as ':'. After adding a
    new line at the start of text the error is solved.
    """
    char_list_to_add_new_line_on_top_of_text = [":"]
    for char in char_list_to_add_new_line_on_top_of_text:
        if char in text:
            return '|' + '\n' + depth * DEPTH_SIZE + text
    return text


# pylint: disable=too-many-instance-attributes
class Nxdl2yaml():
    """
        Parse XML file and print a YML file
    """

    def __init__(
            self,
            symbol_list: List[str],
            root_level_definition: List[str],
            root_level_doc='',
            root_level_symbols=''):

        # updated part of yaml_dict
        self.found_definition = False
        self.root_level_doc = root_level_doc
        self.root_level_symbols = root_level_symbols
        self.root_level_definition = root_level_definition
        self.symbol_list = symbol_list
        self.is_last_element_comment = False
        self.include_comment = True
        self.pi_comments = None
        # NOTE: Here is how root_level_comments organised for storing comments
        # root_level_comment= {'root_doc': comment,
        #                      'symbols': comment,
        #       The 'symbol_doc_comments' list is for comments from all 'symbol doc'
        #                      'symbol_doc_comments' : [comments]
        #                      'symbol_list': [symbols],
        #       The 'symbol_comments' contains comments for 'symbols doc' and all 'symbol'
        #                      'symbol_comments': [comments]}
        self.root_level_comment: Dict[str, str] = {}

    def print_yml(self, input_file, output_yml, verbose):
        """
            Parse an XML file provided as input and print a YML file
        """
        if os.path.isfile(output_yml):
            os.remove(output_yml)

        depth = 0

        self.pi_comments, root = parse(input_file)
        xml_tree = {'tree': root, 'node': root}
        self.xmlparse(output_yml, xml_tree, depth, verbose)

    def handle_symbols(self, depth, node):
        """Handle symbols field and its childs symbol"""

        # pylint: disable=consider-using-f-string
        self.root_level_symbols = (
            f"{remove_namespace_from_tag(node.tag)}: "
            f"{node.text.strip() if node.text else ''}"
        )
        depth += 1
        last_comment = ''
        sbl_doc_cmnt_list = []
        # Comments that come above symbol tag
        symbol_cmnt_list = []
        for child in list(node):
            tag = remove_namespace_from_tag(child.tag)
            if tag == CMNT_TAG and self.include_comment:
                last_comment = self.comvert_to_ymal_comment(depth * DEPTH_SIZE, child.text)
            if tag == 'doc':
                symbol_cmnt_list.append(last_comment)
                # The bellow line is for handling lenth of 'symbol_comments' and
                # 'symbol_doc_comments'. Otherwise print_root_level_info() gets inconsistency
                # over for the loop while writting comment on file
                sbl_doc_cmnt_list.append('')
                last_comment = ''
                self.symbol_list.append(self.handle_not_root_level_doc(depth,
                                                                       text=child.text))
            elif tag == 'symbol':
                # place holder is symbol name
                symbol_cmnt_list.append(last_comment)
                last_comment = ''
                if 'doc' in child.attrib:
                    self.symbol_list.append(
                        self.handle_not_root_level_doc(depth,
                                                       tag=child.attrib['name'],
                                                       text=child.attrib['doc']))
                else:
                    for symbol_doc in list(child):
                        tag = remove_namespace_from_tag(symbol_doc.tag)
                        if tag == CMNT_TAG and self.include_comment:
                            last_comment = self.comvert_to_ymal_comment(depth * DEPTH_SIZE,
                                                                        symbol_doc.text)
                        if tag == 'doc':
                            sbl_doc_cmnt_list.append(last_comment)
                            last_comment = ''
                            self.symbol_list.append(
                                self.handle_not_root_level_doc(depth,
                                                               tag=child.attrib['name'],
                                                               text=symbol_doc.text))
        self.store_root_level_comments('symbol_doc_comments', sbl_doc_cmnt_list)
        self.store_root_level_comments('symbol_comments', symbol_cmnt_list)

    def store_root_level_comments(self, holder, comment):
        """Store yaml text or section line and the comments inteded for that lines or section"""

        self.root_level_comment[holder] = comment

    def handle_definition(self, node):
        """
            Handle definition group and its attributes
            NOTE: Here we tried to store the order of the xml element attributes. So that we get
            exactly the same file in nxdl from yaml.
        """
        # pylint: disable=consider-using-f-string
        # self.root_level_definition[0] = ''
        keyword = ''
        # tmp_word for reseving the location
        tmp_word = "#xx#"
        attribs = node.attrib
        # for tracking the order of name and type
        keyword_order = -1
        for item in attribs:
            if "name" in item:
                keyword = keyword + attribs[item]
                if keyword_order == -1:
                    self.root_level_definition.append(tmp_word)
                    keyword_order = self.root_level_definition.index(tmp_word)
            elif "extends" in item:
                keyword = f"{keyword}({attribs[item]})"
                if keyword_order == -1:
                    self.root_level_definition.append(tmp_word)
                    keyword_order = self.root_level_definition.index(tmp_word)
            elif 'schemaLocation' not in item \
                    and 'extends' != item:
                text = f"{item}: {attribs[item]}"
                self.root_level_definition.append(text)
        self.root_level_definition[keyword_order] = f"{keyword}:"

    def handle_root_level_doc(self, node):
        """
            Handle the documentation field found at root level.
        """
        # tag = remove_namespace_from_tag(node.tag)
        text = node.text
        text = self.handle_not_root_level_doc(depth=0, text=text)
        self.root_level_doc = text

    # pylint: disable=too-many-branches
    def handle_not_root_level_doc(self, depth, text, tag='doc', file_out=None):
        """
        Handle docs field along the yaml file. In this function we also tried to keep
        the track of intended indentation. E.g. the bollow doc block.
            * Topic name
                Description of topic
        """

        # Handling empty doc
        if not text:
            text = ""
        else:
            text = handle_mapping_char(text, -1, True)
        if "\n" in text:
            # To remove '\n' character as it will be added before text.
            text = cleaning_empty_lines(text.split('\n'))
            text_tmp = []
            yaml_indent_n = len((depth + 1) * DEPTH_SIZE)
            # Find indentaion in the first text line with alphabet
            tmp_i = 0
            while tmp_i != -1:
                first_line_indent_n = 0
                # Taking care of empty text whitout any character
                if len(text) == 1 and text[0] == '':
                    break
                for ch_ in text[tmp_i]:
                    if ch_ == ' ':
                        first_line_indent_n = first_line_indent_n + 1
                    elif ch_ != '':
                        tmp_i = -2
                        break
                tmp_i = tmp_i + 1
            # Taking care of doc like bellow:
            # <doc>Text liness
            # text continues</doc>
            # So no indentaion at the staring or doc. So doc group will come along general
            # alignment
            if first_line_indent_n == 0:
                first_line_indent_n = yaml_indent_n

            # for indent_diff -ve all lines will move left by the same ammout
            # for indect_diff +ve all lines will move right the same amount
            indent_diff = yaml_indent_n - first_line_indent_n
            # CHeck for first line empty if not keep first line empty

            for _, line in enumerate(text):
                line_indent_n = 0
                # Collect first empty space without alphabate
                for ch_ in line:
                    if ch_ == ' ':
                        line_indent_n = line_indent_n + 1
                    else:
                        break
                line_indent_n = line_indent_n + indent_diff
                if line_indent_n < yaml_indent_n:
                    # if line still under yaml identation
                    text_tmp.append(yaml_indent_n * ' ' + line.strip())
                else:
                    text_tmp.append(line_indent_n * ' ' + line.strip())

            text = '\n' + '\n'.join(text_tmp)
            if "}" in tag:
                tag = remove_namespace_from_tag(tag)
            indent = depth * DEPTH_SIZE
        elif text:
            text = '\n' + (depth + 1) * DEPTH_SIZE + text.strip()
            if "}" in tag:
                tag = remove_namespace_from_tag(tag)
            indent = depth * DEPTH_SIZE
        else:
            text = ""
            if "}" in tag:
                tag = remove_namespace_from_tag(tag)
            indent = depth * DEPTH_SIZE

        doc_str = f"{indent}{tag}: |{text}\n"
        if file_out:
            file_out.write(doc_str)
            return None
        return doc_str

    def write_out(self, indent, text, file_out):
        """
        Write text line in output file.
        """
        line_string = f"{indent}{text.rstrip()}\n"
        file_out.write(line_string)

    def print_root_level_doc(self, file_out):
        """
        Print at the root level of YML file \
        the general documentation field found in XML file
        """
        indent = 0 * DEPTH_SIZE

        if ('root_doc' in self.root_level_comment
                and self.root_level_comment['root_doc'] != ''):
            text = self.root_level_comment['root_doc']
            self.write_out(indent, text, file_out)

        text = self.root_level_doc
        self.write_out(indent, text, file_out)
        self.root_level_doc = ''

    def comvert_to_ymal_comment(self, indent, text):
        """
            Convert into yaml comment by adding exta '#' char in front of comment lines
        """
        lines = text.split('\n')
        mod_lines = []
        for line in lines:
            line = line.strip()
            if line and line[0] != '#':
                line = indent + '# ' + line
                mod_lines.append(line)
            elif line:
                line = indent + line
                mod_lines.append(line)
        # The starting '\n' to keep multiple comments separate
        return '\n' + '\n'.join(mod_lines)

    def print_root_level_info(self, depth, file_out):
        """
        Print at the root level of YML file \
        the information stored as definition attributes in the XML file
        """
        # pylint: disable=consider-using-f-string
        if depth < 0:
            raise ValueError("Somthing wrong with indentaion in root level.")

        has_categoty = False
        for def_line in self.root_level_definition:
            if def_line in ("category: application", "category: base"):
                self.write_out(indent=0 * DEPTH_SIZE, text=def_line, file_out=file_out)
                # file_out.write(f"{def_line}\n")
                has_categoty = True

        if not has_categoty:
            raise ValueError("Definition dose not get any category from 'base or application'.")
        self.print_root_level_doc(file_out)
        if 'symbols' in self.root_level_comment and self.root_level_comment['symbols'] != '':
            indent = depth * DEPTH_SIZE
            text = self.root_level_comment['symbols']
            self.write_out(indent, text, file_out)
        if self.root_level_symbols:
            self.write_out(indent=0 * DEPTH_SIZE, text=self.root_level_symbols, file_out=file_out)
            # symbol_list include 'symbols doc', and all 'symbol'
            for ind, symbol in enumerate(self.symbol_list):
                # Taking care of comments that come on to of 'symbols doc' and 'symbol'
                if 'symbol_comments' in self.root_level_comment and \
                        self.root_level_comment['symbol_comments'][ind] != '':
                    indent = depth * DEPTH_SIZE
                    self.write_out(indent,
                                   self.root_level_comment['symbol_comments'][ind], file_out)
                if 'symbol_doc_comments' in self.root_level_comment and \
                        self.root_level_comment['symbol_doc_comments'][ind] != '':

                    indent = depth * DEPTH_SIZE
                    self.write_out(indent,
                                   self.root_level_comment['symbol_doc_comments'][ind], file_out)

                self.write_out(indent=(0 * DEPTH_SIZE), text=symbol, file_out=file_out)
        if len(self.pi_comments) > 1:
            indent = DEPTH_SIZE * depth
            # The first comment is top level copy-right doc string
            for comment in self.pi_comments[1:]:
                self.write_out(indent, self.comvert_to_ymal_comment(indent, comment), file_out)
        if self.root_level_definition:
            # Soring NXname for writting end of the definition attributes
            nx_name = ''
            for defs in self.root_level_definition:
                if 'NX' in defs and defs[-1] == ':':
                    nx_name = defs
                    continue
                if defs in ("category: application", "category: base"):
                    continue
                self.write_out(indent=0 * DEPTH_SIZE, text=defs, file_out=file_out)
            self.write_out(indent=0 * DEPTH_SIZE, text=nx_name, file_out=file_out)
        self.found_definition = False

    def handle_exists(self, exists_dict, key, val):
        """
            Create exist component as folows:

            {'min' : value for min,
             'max' : value for max,
             'optional' : value for optional}

            This is created separately so that the keys stays in order.
        """
        if not val:
            val = ''
        else:
            val = str(val)
        if 'minOccurs' == key:
            exists_dict['minOccurs'] = ['min', val]
        if 'maxOccurs' == key:
            exists_dict['maxOccurs'] = ['max', val]
        if 'optional' == key:
            exists_dict['optional'] = ['optional', val]
        if 'recommended' == key:
            exists_dict['recommended'] = ['recommended', val]
        if 'required' == key:
            exists_dict['required'] = ['required', val]

    # pylint: disable=too-many-branches, consider-using-f-string
    def handle_group_or_field(self, depth, node, file_out):
        """Handle all the possible attributes that come along a field or group"""

        allowed_attr = ['optional', 'recommended', 'name', 'type', 'axes', 'axis', 'data_offset',
                        'interpretation', 'long_name', 'maxOccurs', 'minOccurs', 'nameType',
                        'optional', 'primary', 'signal', 'stride', 'units', 'required',
                        'deprecated', 'exists']

        name_type = ""
        node_attr = node.attrib
        rm_key_list = []
        # Maintain order: name and type in form name(type) or (type)name that come first
        for key, val in node_attr.items():
            if key == 'name':
                name_type = name_type + val
                rm_key_list.append(key)
            if key == 'type':
                name_type = name_type + "(%s)" % val
                rm_key_list.append(key)
        if not name_type:
            raise ValueError(f"No 'name' or 'type' hase been found. But, 'group' or 'field' "
                             f"must have at list a nme.We got attributes:  {node_attr}")
        file_out.write('{indent}{name_type}:\n'.format(
            indent=depth * DEPTH_SIZE,
            name_type=name_type))

        for key in rm_key_list:
            del node_attr[key]

        # tmp_dict intended to persevere order of attribnutes
        tmp_dict = {}
        exists_dict = {}
        for key, val in node_attr.items():
            # As both 'minOccurs', 'maxOccurs' and optionality move to the 'exists'
            if key in ['minOccurs', 'maxOccurs', 'optional', 'recommended', 'required']:
                if 'exists' not in tmp_dict:
                    tmp_dict['exists'] = []
                self.handle_exists(exists_dict, key, val)
            elif key == 'units':
                tmp_dict['unit'] = str(val)
            else:
                tmp_dict[key] = str(val)
            if key not in allowed_attr:
                raise ValueError(f"An attribute ({key}) in 'field' or 'group' has been found "
                                 f"that is not allowed. The allowed attr is {allowed_attr}.")

        if exists_dict:
            for key, val in exists_dict.items():
                if key in ['minOccurs', 'maxOccurs']:
                    tmp_dict['exists'] = tmp_dict['exists'] + val
                elif key in ['optional', 'recommended', 'required']:
                    tmp_dict['exists'] = key

        depth_ = depth + 1
        for key, val in tmp_dict.items():
            # Increase depth size inside handle_map...() for writting text with one
            # more indentation.
            file_out.write(f'{depth_ * DEPTH_SIZE}{key}: '
                           f'{handle_mapping_char(val, depth_ + 1, False)}\n')

    # pylint: disable=too-many-branches, too-many-locals
    def handle_dimension(self, depth, node, file_out):
        """
        Handle the dimension field.
            NOTE: Usually we take care of any xml element in xmlparse(...) and
        recursion_in_xml_tree(...) functions. But Here it is a bit different. The doc dimension
          and attributes of dim has been handled inside this function here.
        """
        # pylint: disable=consider-using-f-string
        possible_dim_attrs = ['ref', 'required',
                              'incr', 'refindex']
        possible_dimemsion_attrs = ['rank']

        # taking care of Dimension tag
        file_out.write(
            '{indent}{tag}:\n'.format(
                indent=depth * DEPTH_SIZE,
                tag=node.tag.split("}", 1)[1]))
        # Taking care of dimension attributes
        for attr, value in node.attrib.items():
            if attr in possible_dimemsion_attrs and not isinstance(value, dict):
                indent = (depth + 1) * DEPTH_SIZE
                file_out.write(f'{indent}{attr}: {value}\n')
            else:
                raise ValueError(f"Dimension has got an attribute {attr} that is not valid."
                                 f"Current the allowd atributes are {possible_dimemsion_attrs}."
                                 f" Please have a look")
        # taking carew of dimension doc
        for child in list(node):
            tag = remove_namespace_from_tag(child.tag)
            if tag == 'doc':
                text = self.handle_not_root_level_doc(depth + 1, child.text)
                file_out.write(text)
                node.remove(child)

        dim_index_value = ''
        dim_other_parts = {}
        dim_cmnt_node = []
        # taking care of dim and doc childs of dimension
        for child in list(node):
            tag = remove_namespace_from_tag(child.tag)
            child_attrs = child.attrib
            # taking care of index and value attributes
            if tag == ('dim'):
                # taking care of index and value in format [[index, value]]
                dim_index_value = dim_index_value + '[{index}, {value}], '.format(
                    index=child_attrs['index'] if "index" in child_attrs else '',
                    value=child_attrs['value'] if "value" in child_attrs else '')
                if "index" in child_attrs:
                    del child_attrs["index"]
                if "value" in child_attrs:
                    del child_attrs["value"]

                # Taking care of doc comes as child of dim
                for cchild in list(child):
                    ttag = cchild.tag.split("}", 1)[1]
                    if ttag == ('doc'):
                        if ttag not in dim_other_parts:
                            dim_other_parts[ttag] = []
                        text = cchild.text
                        dim_other_parts[ttag].append(text.strip())
                        child.remove(cchild)
                        continue
                # taking care of other attributes except index and value
                for attr, value in child_attrs.items():
                    if attr in possible_dim_attrs:
                        if attr not in dim_other_parts:
                            dim_other_parts[attr] = []
                        dim_other_parts[attr].append(value)
            if tag == CMNT_TAG and self.include_comment:
                # Store and remove node so that comment nodes from dim node so
                # that it does not call in xmlparser function
                dim_cmnt_node.append(child)
                node.remove(child)

        # All 'dim' element comments on top of 'dim' yaml key
        if dim_cmnt_node:
            for ch_nd in dim_cmnt_node:
                self.handel_comment(depth + 1, ch_nd, file_out)
        # index and value attributes of dim elements
        file_out.write(
            '{indent}dim: [{value}]\n'.format(
                indent=(depth + 1) * DEPTH_SIZE,
                value=dim_index_value[:-2] or ''))
        # Write the attributes, except index and value, and doc of dim as child of dim_parameter.
        # But tthe doc or attributes for each dim come inside list according to the order of dim.
        if dim_other_parts:
            file_out.write(
                '{indent}dim_parameters:\n'.format(
                    indent=(depth + 1) * DEPTH_SIZE))
            # depth = depth + 2 dim_paramerter has child such as doc of dim
            indent = (depth + 2) * DEPTH_SIZE
            for key, value in dim_other_parts.items():
                if key == 'doc':
                    value = self.handle_not_root_level_doc(depth + 2, str(value), key, file_out)
                else:
                    # Increase depth size inside handle_map...() for writting text with one
                    # more indentation.
                    file_out.write(f"{indent}{key}: "
                                   f"{handle_mapping_char(value, depth + 3, False)}\n")

    def handle_enumeration(self, depth, node, file_out):
        """
            Handle the enumeration field parsed from the xml file.

        If the enumeration items contain a doc field, the yaml file will contain items as child
        fields of the enumeration field.

        If no doc are inherited in the enumeration items, a list of the items is given for the
        enumeration list.

    """
        # pylint: disable=consider-using-f-string

        check_doc = []
        for child in list(node):
            if list(child):
                check_doc.append(list(child))
        # pylint: disable=too-many-nested-blocks
        if check_doc:
            file_out.write(
                '{indent}{tag}: \n'.format(
                    indent=depth * DEPTH_SIZE,
                    tag=node.tag.split("}", 1)[1]))
            for child in list(node):
                tag = remove_namespace_from_tag(child.tag)
                itm_depth = depth + 1
                if tag == ('item'):
                    file_out.write(
                        '{indent}{value}: \n'.format(
                            indent=(itm_depth) * DEPTH_SIZE,
                            value=child.attrib['value']))

                    if list(child):
                        for item_doc in list(child):
                            if remove_namespace_from_tag(item_doc.tag) == 'doc':
                                item_doc_depth = itm_depth + 1
                                self.handle_not_root_level_doc(item_doc_depth, item_doc.text,
                                                               item_doc.tag, file_out)
                            if (remove_namespace_from_tag(item_doc.tag) == CMNT_TAG
                                    and self.include_comment):
                                self.handel_comment(itm_depth + 1, item_doc, file_out)
                if tag == CMNT_TAG and self.include_comment:
                    self.handel_comment(itm_depth + 1, child, file_out)
        else:
            enum_list = ''
            remove_nodes = []
            for item_child in list(node):
                tag = remove_namespace_from_tag(item_child.tag)
                if tag == ('item'):
                    enum_list = enum_list + '{value}, '.format(
                        value=item_child.attrib['value'])
                if tag == CMNT_TAG and self.include_comment:
                    self.handel_comment(depth, item_child, file_out)
                    remove_nodes.append(item_child)
            for ch_node in remove_nodes:
                node.remove(ch_node)

            file_out.write(
                '{indent}{tag}: [{enum_list}]\n'.format(
                    indent=depth * DEPTH_SIZE,
                    tag=remove_namespace_from_tag(node.tag),
                    enum_list=enum_list[:-2] or ''))

    def handle_attributes(self, depth, node, file_out):
        """Handle the attributes parsed from the xml file"""

        allowed_attr = ['name', 'type', 'units', 'nameType', 'recommended', 'optional',
                        'minOccurs', 'maxOccurs', 'deprecated']

        name = ""
        node_attr = node.attrib
        if 'name' in node_attr:
            pass
        else:
            raise ValueError("Attribute must have an name key.")
        rm_key_list = []
        # Maintain order: name and type in form name(type) or (type)name that come first
        for key, val in node_attr.items():
            if key == 'name':
                name = val
                rm_key_list.append(key)

        for key in rm_key_list:
            del node_attr[key]

        file_out.write('{indent}{escapesymbol}{name}:\n'.format(
            indent=depth * DEPTH_SIZE,
            escapesymbol=r'\@',
            name=name))

        tmp_dict = {}
        exists_dict = {}
        for key, val in node_attr.items():
            # As both 'minOccurs', 'maxOccurs' and optionality move to the 'exists'
            if key in ['minOccurs', 'maxOccurs', 'optional', 'recommended', 'required']:
                if 'exists' not in tmp_dict:
                    tmp_dict['exists'] = []
                self.handle_exists(exists_dict, key, val)
            elif key == 'units':
                tmp_dict['unit'] = val
            else:
                tmp_dict[key] = val
            if key not in allowed_attr:
                raise ValueError(f"An attribute ({key}) has been found that is not allowed."
                                 f"The allowed attr is {allowed_attr}.")

        has_min_max = False
        has_opt_reco_requ = False
        if exists_dict:
            for key, val in exists_dict.items():
                if key in ['minOccurs', 'maxOccurs']:
                    tmp_dict['exists'] = tmp_dict['exists'] + val
                    has_min_max = True
                elif key in ['optional', 'recommended', 'required']:
                    tmp_dict['exists'] = key
                    has_opt_reco_requ = True
        if has_min_max and has_opt_reco_requ:
            raise ValueError("Optionality 'exists' can take only either from ['minOccurs',"
                             " 'maxOccurs'] or from ['optional', 'recommended', 'required']"
                             ". But not from both of the groups together. Please check in"
                             " attributes")

        depth_ = depth + 1
        for key, val in tmp_dict.items():
            # Increase depth size inside handle_map...() for writting text with one
            # more indentation.
            file_out.write(f'{depth_ * DEPTH_SIZE}{key}: '
                           f'{handle_mapping_char(val, depth_ + 1, False)}\n')

    def handel_link(self, depth, node, file_out):
        """
            Handle link elements of nxdl
        """

        possible_link_attrs = ['name', 'target', 'napimount']
        node_attr = node.attrib
        # Handle special cases
        if 'name' in node_attr:
            file_out.write('{indent}{name}(link):\n'.format(
                indent=depth * DEPTH_SIZE,
                name=node_attr['name'] or ''))
            del node_attr['name']

        depth_ = depth + 1
        # Handle general cases
        for attr_key, val in node_attr.items():
            if attr_key in possible_link_attrs:
                file_out.write('{indent}{attr}: {value}\n'.format(
                    indent=depth_ * DEPTH_SIZE,
                    attr=attr_key,
                    value=val))
            else:
                raise ValueError(f"An anexpected attribute '{attr_key}' of link has found."
                                 f"At this moment the alloed keys are {possible_link_attrs}")

    def handel_choice(self, depth, node, file_out):
        """
            Handle choice element which is a parent node of group.
        """

        possible_attr = []

        node_attr = node.attrib
        # Handle special casees
        if 'name' in node_attr:
            file_out.write('{indent}{attr}(choice): \n'.format(
                indent=depth * DEPTH_SIZE,
                attr=node_attr['name']))
            del node_attr['name']

        depth_ = depth + 1
        # Taking care of general attrinutes. Though, still no attrinutes have found,
        # but could be used for future
        for attr in node_attr.items():
            if attr in possible_attr:
                file_out.write('{indent}{attr}: {value}\n'.format(
                    indent=depth_ * DEPTH_SIZE,
                    attr=attr,
                    value=node_attr[attr]))
            else:
                raise ValueError(f"An unexpected attribute '{attr}' of 'choice' has been found."
                                 f"At this moment attributes for choice {possible_attr}")

    def handel_comment(self, depth, node, file_out):
        """
            Collect comment element and pass to write_out function
        """
        indent = depth * DEPTH_SIZE
        if self.is_last_element_comment:
            text = self.comvert_to_ymal_comment(indent, node.text)
            self.write_out(indent, text, file_out)
        else:
            text = self.comvert_to_ymal_comment(indent, node.text)
            self.write_out(indent, text, file_out)
            self.is_last_element_comment = True

    def recursion_in_xml_tree(self, depth, xml_tree, output_yml, verbose):
        """
            Descend lower level in xml tree. If we are in the symbols branch, the recursive
        behaviour is not triggered as we already handled the symbols' childs.
        """

        tree = xml_tree['tree']
        node = xml_tree['node']
        for child in list(node):
            xml_tree_children = {'tree': tree, 'node': child}
            self.xmlparse(output_yml, xml_tree_children, depth, verbose)

    # pylint: disable=too-many-branches, too-many-statements
    def xmlparse(self, output_yml, xml_tree, depth, verbose):
        """
        Main of the nxdl2yaml converter.
        It parses XML tree, then prints recursively each level of the tree
        """
        tree = xml_tree['tree']
        node = xml_tree['node']
        if verbose:
            sys.stdout.write(f'Node tag: {remove_namespace_from_tag(node.tag)}\n')
            sys.stdout.write(f'Attributes: {node.attrib}\n')
        with open(output_yml, "a", encoding="utf-8") as file_out:
            tag = remove_namespace_from_tag(node.tag)
            if tag == 'definition':
                self.found_definition = True
                self.handle_definition(node)
                # Taking care of root level doc and symbols
                remove_cmnt_n = None
                last_comment = ''
                for child in list(node):
                    tag_tmp = remove_namespace_from_tag(child.tag)
                    if tag_tmp == CMNT_TAG and self.include_comment:
                        last_comment = self.comvert_to_ymal_comment(depth * DEPTH_SIZE, child.text)
                        remove_cmnt_n = child
                    if tag_tmp == 'doc':
                        self.store_root_level_comments('root_doc', last_comment)
                        last_comment = ''
                        self.handle_root_level_doc(child)
                        node.remove(child)
                        if remove_cmnt_n is not None:
                            node.remove(remove_cmnt_n)
                            remove_cmnt_n = None
                    if tag_tmp == 'symbols':
                        self.store_root_level_comments('symbols', last_comment)
                        last_comment = ''
                        self.handle_symbols(depth, child)
                        node.remove(child)
                        if remove_cmnt_n is not None:
                            node.remove(remove_cmnt_n)
                            remove_cmnt_n = None

            if tag == ('doc') and depth != 1:
                parent = get_node_parent_info(tree, node)[0]
                doc_parent = remove_namespace_from_tag(parent.tag)
                if doc_parent != 'item':
                    self.handle_not_root_level_doc(depth, text=node.text,
                                                   tag=node.tag,
                                                   file_out=file_out)

            if self.found_definition is True and self.root_level_doc:
                self.print_root_level_info(depth, file_out)
            # End of print root-level definitions in file
            if tag in ('field', 'group') and depth != 0:
                self.handle_group_or_field(depth, node, file_out)
            if tag == ('enumeration'):
                self.handle_enumeration(depth, node, file_out)
            if tag == ('attribute'):
                self.handle_attributes(depth, node, file_out)
            if tag == ('dimensions'):
                self.handle_dimension(depth, node, file_out)
            if tag == ('link'):
                self.handel_link(depth, node, file_out)
            if tag == ('choice'):
                self.handel_choice(depth, node, file_out)
            if tag == CMNT_TAG and self.include_comment:
                self.handel_comment(depth, node, file_out)
        depth += 1
        # Write nested nodes
        self.recursion_in_xml_tree(depth, xml_tree, output_yml, verbose)


def compare_niac_and_my(tree, tree2, verbose, node, root_no_duplicates):
    """This function creates two trees with Niac XML file and My XML file.
The main aim is to compare the two trees and create a new one that is the
union of the two initial trees.

"""
    root = tree.getroot()
    root2 = tree2.getroot()
    attrs_list_niac = []
    for nodo in root.iter(node):
        attrs_list_niac.append(nodo.attrib)
    if verbose:
        sys.stdout.write('Attributes found in Niac file: \n')
        sys.stdout.write(str(attrs_list_niac) + '\n')
        sys.stdout.write('  \n')
        sys.stdout.write('Started merging of Niac and My file... \n')
    for elem in root.iter(node):
        if verbose:
            sys.stdout.write('- Niac element inserted: \n')
            sys.stdout.write(str(elem.attrib) + '\n')
        index = get_node_parent_info(tree, elem)[1]
        root_no_duplicates.insert(index, elem)

    for elem2 in root2.iter(node):
        index = get_node_parent_info(tree2, elem2)[1]
        if elem2.attrib not in attrs_list_niac:
            if verbose:
                sys.stdout.write('- My element inserted: \n')
                sys.stdout.write(str(elem2.attrib) + '\n')
            root_no_duplicates.insert(index, elem2)

    if verbose:
        sys.stdout.write('     \n')
    return root_no_duplicates



"""
Tests the version retrieval for the nexus definitions submodule
"""
import re

from pynxtools import get_nexus_version


def test_get_nexus_version():
    """
    Tests if we get a version string from nexus definitions
    """
    version = get_nexus_version()

    assert version is not None
    assert re.match(r"v\d{4}\.\d{2}\.post1\.dev\d+\+g[a-z0-9]", version)



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Implementation for refractiveindex.info dispersion functions"""
from io import StringIO
from dataclasses import dataclass
from typing import Callable, Dict, List, Optional
from numpy import sqrt
import pandas as pd

ModelNameCallback = Callable[[str, str], None]
ModelInfoCallback = Callable[["DispersionInfo", "ModelInput"], str]
SingleParamCallback = Callable[[str, str, float, str], None]
RepeatedParamCallback = Callable[[str, str, List[float], List[str]], None]
TableCallback = Callable[[str, str, pd.DataFrame], None]


@dataclass
class TemplateCallbacks:
    """Contains callbacks to write model infos into the nexus template"""

    add_model_name: ModelNameCallback
    add_model_info: ModelInfoCallback
    add_single_param: SingleParamCallback
    add_rep_param: RepeatedParamCallback


@dataclass
class ModelInput:
    """The input data and callbacks for the model functions"""

    dispersion_path: str
    coeffs: List[float]
    convert_to_n: bool
    callbacks: TemplateCallbacks
    wavelength_range: Optional[str] = None


@dataclass
class DispersionInfo:
    """Holds information about a dispersion"""

    name: str
    representation: str
    formula: str


@dataclass
class TableInput:
    """The input data and callbacks for tabular dispersions"""

    dispersion_path: str
    nx_name: str
    table_type: str
    data: str
    add_table: TableCallback


def tabulated(table_input: TableInput):
    """
    Parsing for tabulated dispersions from refractiveindex.info
    """
    index_from = {
        "tabulated nk": ["Wavelength", "n", "k"],
        "tabulated n": ["Wavelength", "n"],
        "tabulated k": ["Wavelength", "k"],
    }

    refractive_index = pd.read_table(
        StringIO(table_input.data), sep="\\s+", names=index_from[table_input.table_type]
    )
    refractive_index.set_index("Wavelength", inplace=True)

    refractive_index["refractive_index"] = 0 + 0j
    if "n" in refractive_index:
        refractive_index.loc[:, "refractive_index"] += refractive_index["n"]
    if "k" in refractive_index:
        refractive_index.loc[:, "refractive_index"] += 1j * refractive_index["k"]

    table_input.add_table(
        table_input.dispersion_path, table_input.nx_name, refractive_index
    )


def sellmeier_squared(model_input: ModelInput):
    """
    The Sellmeier formula (1) from refractiveindex.info
    See https://refractiveindex.info/database/doc/Dispersion%20formulas.pdf
    for a reference.
    """
    model_input.callbacks.add_model_name(model_input.dispersion_path, "Sellmeier")
    path = model_input.callbacks.add_model_info(
        DispersionInfo(
            "Sellmeier",
            "eps",
            "eps_inf + 1 + sum[A * lambda ** 2 / (lambda ** 2 - B ** 2)]",
        ),
        model_input,
    )

    model_input.callbacks.add_single_param(path, "eps_inf", model_input.coeffs[0], "")
    model_input.callbacks.add_rep_param(path, "A", model_input.coeffs[1::2], [""])
    model_input.callbacks.add_rep_param(
        path, "B", model_input.coeffs[2::2], ["micrometer"]
    )


def sellmeier(model_input: ModelInput):
    """
    The Sellmeier-2 formula (2) from refractiveindex.info
    See https://refractiveindex.info/database/doc/Dispersion%20formulas.pdf
    for a reference.
    """
    for i in range(2, len(model_input.coeffs), 2):
        model_input.coeffs[i] = sqrt(model_input.coeffs[i])
    sellmeier_squared(model_input)


def polynomial(model_input: ModelInput):
    """
    The Polynomial formula (3) from refractiveindex.info
    See https://refractiveindex.info/database/doc/Dispersion%20formulas.pdf
    for a reference.
    """
    model_input.callbacks.add_model_name(model_input.dispersion_path, "Polynomial")
    path = model_input.callbacks.add_model_info(
        DispersionInfo("Polynomial", "eps", "eps_inf + sum[f * lambda ** e]"),
        model_input,
    )

    model_input.callbacks.add_single_param(path, "eps_inf", model_input.coeffs[0], "")
    model_input.callbacks.add_rep_param(
        path,
        "f",
        model_input.coeffs[1::2],
        [f"1/micrometer^{e}" for e in model_input.coeffs[2::2]],
    )
    model_input.callbacks.add_rep_param(path, "e", model_input.coeffs[2::2], [""])


def sellmeier_polynomial(model_input: ModelInput):
    """
    The Sellmeier Polynomial formula (4) from refractiveindex.info
    See https://refractiveindex.info/database/doc/Dispersion%20formulas.pdf
    for a reference.
    """
    model_input.callbacks.add_model_name(model_input.dispersion_path, "Sellmeier")

    a_units = []
    # Check for valid parameters
    for exponent1 in model_input.coeffs[2:7:4]:
        if exponent1 not in [0, 2]:
            raise ValueError(f"e1 may only be 0 or 1 but is {exponent1}")
        a_units.append("" if exponent1 == 0 else "1/micrometer^2")

    path = model_input.callbacks.add_model_info(
        DispersionInfo(
            "Sellmeier",
            "eps",
            "eps_inf + sum[A * lambda ** e1 / (lambda ** 2 - B**e2)]",
        ),
        model_input,
    )

    model_input.callbacks.add_single_param(path, "eps_inf", model_input.coeffs[0], "")
    model_input.callbacks.add_rep_param(path, "A", model_input.coeffs[1:6:4], a_units)
    model_input.callbacks.add_rep_param(
        path,
        "B",
        model_input.coeffs[3:8:4],
        [f"micrometer^(2/{e2})" for e2 in model_input.coeffs[4:9:4]],
    )
    model_input.callbacks.add_rep_param(path, "e1", model_input.coeffs[2:7:4], [""])
    model_input.callbacks.add_rep_param(path, "e2", model_input.coeffs[4:9:4], [""])

    # If there are just 9 parameters don't append polynomial part
    if len(model_input.coeffs) <= 9:
        return

    model_input.callbacks.add_model_name(
        model_input.dispersion_path,
        "Polynomial + Sellmeier (from RefractiveIndex.Info)",
    )
    path = model_input.callbacks.add_model_info(
        DispersionInfo("Polynomial", "eps", "sum[f * lambda ** e]"),
        model_input,
    )

    model_input.callbacks.add_rep_param(
        path,
        "f",
        model_input.coeffs[9::2],
        [f"1/micrometer^{e}" for e in model_input.coeffs[10::2]],
    )
    model_input.callbacks.add_rep_param(path, "e", model_input.coeffs[10::2], [""])


def cauchy(model_input: ModelInput):
    """
    The Cauchy formula (5) from refractiveindex.info
    See https://refractiveindex.info/database/doc/Dispersion%20formulas.pdf
    for a reference.
    """
    model_input.callbacks.add_model_name(model_input.dispersion_path, "Cauchy")
    path = model_input.callbacks.add_model_info(
        DispersionInfo("Cauchy", "n", "n_inf + sum[f * lambda ** e]"), model_input
    )

    model_input.callbacks.add_single_param(path, "n_inf", model_input.coeffs[0], "")
    model_input.callbacks.add_rep_param(
        path,
        "f",
        model_input.coeffs[1::2],
        [f"1/micrometer^{e}" for e in model_input.coeffs[2::2]],
    )
    model_input.callbacks.add_rep_param(path, "e", model_input.coeffs[2::2], [""])


def gases(model_input: ModelInput):
    """
    The Gases dispersion formula (6) from refractiveindex.info
    See https://refractiveindex.info/database/doc/Dispersion%20formulas.pdf
    for a reference.
    """
    model_input.callbacks.add_model_name(model_input.dispersion_path, "Gases")
    path = model_input.callbacks.add_model_info(
        DispersionInfo("Gases", "n", "1 + n_inf + sum[A / (B - lambda ** (-2))]"),
        model_input,
    )

    model_input.callbacks.add_single_param(path, "n_inf", model_input.coeffs[0], "")
    model_input.callbacks.add_rep_param(
        path, "A", model_input.coeffs[1::2], ["1/micrometer^2"]
    )
    model_input.callbacks.add_rep_param(
        path, "B", model_input.coeffs[2::2], ["1/micrometer^2"]
    )


def herzberger(model_input: ModelInput):
    """
    The Herzberger dispersion formula (6) from refractiveindex.info
    See https://refractiveindex.info/database/doc/Dispersion%20formulas.pdf
    for a reference.
    """
    model_input.callbacks.add_model_name(model_input.dispersion_path, "Herzberger")
    path = model_input.callbacks.add_model_info(
        DispersionInfo(
            "Herzberger",
            "n",
            "n_inf + C2 / (lambda ** 2 - B1) + "
            "C3 * (lambda ** 2 - B2) ** (-2) + "
            "sum[f * lambda ** e]",
        ),
        model_input,
    )

    f_coeffs = model_input.coeffs[3:6] + [0.0] * (3 - len(model_input.coeffs[3:6]))

    model_input.callbacks.add_single_param(path, "n_inf", model_input.coeffs[0], "")
    model_input.callbacks.add_single_param(
        path, "C2", model_input.coeffs[1], "micrometer^2"
    )
    model_input.callbacks.add_single_param(
        path, "C3", model_input.coeffs[2], "micrometer^4"
    )
    model_input.callbacks.add_single_param(path, "B1", 0.028, "micrometer^2")
    model_input.callbacks.add_single_param(path, "B2", 0.028, "micrometer^2")
    model_input.callbacks.add_rep_param(
        path, "f", f_coeffs, [f"1/micrometer^{e}" for e in [2, 4, 6]]
    )
    model_input.callbacks.add_rep_param(path, "e", [2, 4, 6], [""])


def retro(model_input: ModelInput):
    """
    The exotic formula (8) from refractiveindex.info
    See https://refractiveindex.info/database/doc/Dispersion%20formulas.pdf
    for a reference.
    """
    model_input.callbacks.add_model_name(model_input.dispersion_path, "Retro")
    path = model_input.callbacks.add_model_info(
        DispersionInfo(
            "Retro",
            "eps",
            "(lambda ** 2 * (2 * A + 2 * B + 2 * D * lambda ** 2 + 1) "
            "- C * (2 * A + 2 * D * lambda ** 2 + 1)) "
            "/ (C * (A + D * lambda ** 2 - 1) - "
            "lambda ** 2 * (A + B + D * lambda ** 2 - 1))",
        ),
        model_input,
    )

    names = ["eps_inf", "A", "B", "C", "D"]
    units = ["", "", "", "micrometer^2", "micrometer^(-2)"]

    model_input.coeffs += [0.0] * (len(names) - len(model_input.coeffs))

    for name, coeff, unit in zip(names, model_input.coeffs, units):
        model_input.callbacks.add_single_param(path, name, coeff, unit)


def exotic(model_input: ModelInput):
    """
    The exotic formula (9) from refractiveindex.info
    See https://refractiveindex.info/database/doc/Dispersion%20formulas.pdf
    for a reference.
    """
    model_input.callbacks.add_model_name(model_input.dispersion_path, "Exotic")
    path = model_input.callbacks.add_model_info(
        DispersionInfo(
            "Exotic",
            "eps",
            "eps_inf + C2 / (lambda ** 2 - C3) + "
            "C4 * (lambda - C5) / ((lambda - C5) ** 2 + C6)",
        ),
        model_input,
    )

    names = ["eps_inf"] + [f"C{i}" for i in range(2, 7)]
    units = ["", "micrometer^2", "micrometer^2", "", "micrometer", "micrometer^2"]
    model_input.coeffs += [0.0] * (len(names) - len(model_input.coeffs))

    for name, coeff, unit in zip(names, model_input.coeffs, units):
        model_input.callbacks.add_single_param(path, name, coeff, unit)


SUPPORTED_TABULAR_PARSERS = [f"tabulated {t}" for t in ["nk", "n", "k"]]


FORMULA_PARSERS: Dict[str, Callable[[ModelInput], None]] = {
    "formula 1": sellmeier_squared,
    "formula 2": sellmeier,
    "formula 3": polynomial,
    "formula 4": sellmeier_polynomial,
    "formula 5": cauchy,
    "formula 6": gases,
    "formula 7": herzberger,
    "formula 8": retro,
    "formula 9": exotic,
}



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Parse Kikuchi pattern from a ZIP file without temporary copies of unpacked files."""

# pylint: disable=no-member,duplicate-code

# https://orix.readthedocs.io/en/stable/tutorials/inverse_pole_figures.html

# type: ignore

# from typing import Dict, Any, List

import re

import numpy as np

from zipfile37 import ZipFile

# import imageio.v3 as iio
from PIL import Image as pil

from pynxtools.dataconverter.readers.em_om.utils.em_nexus_plots import HFIVE_WEB_MAX_SIZE


class NxEmOmZipEbsdParser:
    """Parse *.zip EBSD data.

    """

    def __init__(self, file_name, entry_id):
        """Class wrapping zip parser."""
        # this is one example which should be extended to support reading of
        # EBSD pattern simulations and/or generic EBSD pattern in a zip file
        # frequently pattern are stored as individual images and then compressed
        # and uploaded to online repositories like zenodo
        # this rips these data out of their context, instead one should better
        # create an application definition which interfaces how a computer simulation
        # exports these pattern into a common format (e.g. some of the HDF5 variants)
        self.file_name = file_name
        self.entry_id = entry_id
        self.stack_meta = {}
        self.stack = []

    def parse_zip(self, template: dict) -> dict:
        """Parse content from *.zip format."""
        with ZipFile(self.file_name) as zip_file_hdl:
            # print(zip_file_hdl.namelist())
            # identify first the content in this archive
            # ASSUME that pattern have numeral components in their file name
            zip_content_table = {}
            for file in zip_file_hdl.namelist():
                keyword = str(np.uint64(re.sub('[^0-9]', '', file)))
                if len(keyword) > 0 and keyword not in zip_content_table:
                    zip_content_table[keyword] = file
                else:
                    print("WARNING::Automatically derived name is not unique!")
                    return template
            # not all content though qualifies as Kikuchi pattern
            # people can trick an algorithm always
            #     like camouflage an image file for a text file
            #     upload images which have not the same dimensions
            #     upload images which are for sure not Kikuchi pattern and in fact
            # the last point is highly relevant for NOMAD assume a public instance
            # one could easily inject here inappropriate or illegal material...

            # ASSUME first image specifies file type and dimensions for all pattern to come
            # ...here we immediately see how problematic custom directory structures
            # for storing research data are even if they were to contain only exactly
            # always data of expected format...
            self.stack_meta = {"fname": "",
                               "size": (0, 0),
                               "dtype": np.uint8,
                               "ftype": ""}
            # in pixel, use axisy and axisx for dimension scale axes
            # ASSUME slow axis is y, fast axis is x
            for keyword, value in zip_content_table.items():
                tmp = value.split(".")
                if (len(tmp) > 1) and (tmp[-1].lower() in ["bmp", "jpg", "png", "tiff"]):
                    # there are examples where people store Kikuchi diffraction pattern
                    # as lossy and lossless raster...
                    # pil supports reading of files in more formats but the above are
                    # the formats typically used for Kikuchi pattern
                    with zip_file_hdl.open(value) as file_hdl:
                        # ##MK::would be to not load the entire image but just shape tags
                        img = pil.open(file_hdl, "r")  # np.asarray(pil.open(file_hdl))
                        if img.mode != "L":
                            # https://stackoverflow.com/questions/1996577/
                            # how-can-i-get-the-depth-of-a-jpg-file
                            break
                        shp = (img.height, img.width)  # np.shape(img)
                        if (shp[0] > 0) and (shp[0] <= HFIVE_WEB_MAX_SIZE) \
                                and (shp[1] > 0) and (shp[1] <= HFIVE_WEB_MAX_SIZE):
                            # found the guiding image
                            self.stack_meta["size"] = (shp[0], shp[1])  # , 3)
                            self.stack_meta["fname"] = value
                            self.stack_meta["ftype"] = tmp[-1].lower()
                            self.stack_meta["dtype"] = np.uint8  # for mode "L"
                            # ##MK::bit depth and other constraints
                            break
                            # ##MK::at the moment not supporting > HFIVE_WEB_MAX_SIZE
                            # images but this is usually not required
            print(self.stack_meta)
            # ASSUME that the user has stored all other relevant images of the stack
            # with useful numeral names and that these have the same metadata
            # (size, filetype, dtype)
            identifier = 0
            self.stack = np.zeros((len(zip_content_table),
                                   self.stack_meta["size"][0],
                                   self.stack_meta["size"][1]),
                                  self.stack_meta["dtype"])
            for keyword, value in zip_content_table.items():
                tmp = value.split(".")
                if (len(tmp) > 1) and (tmp[-1].lower() == self.stack_meta["ftype"]):
                    with zip_file_hdl.open(value) as file_handle:
                        # img = np.asarray(pil.open(file_handle))
                        img = pil.open(file_handle, "r")
                        if (img.mode == "L") and (img.palette is None):
                            img = np.asarray(img, np.uint8)
                            # ##MK::for bitmap no need to discard alpha channels, check
                            # print(np.shape(img))
                            # img = img[:, :, 0:3]  # discard eventual alpha channels
                        else:
                            break

                        if (np.shape(img) == self.stack_meta["size"]) \
                                and (img.dtype == self.stack_meta["dtype"]):
                            self.stack[identifier, :, :] = img
                            # Kikuchi pattern may come as 8-bit (grayscale) RGBs
                            # or as simulated intensities (as floats)
                            # ASSUME here we load only images when [1, HFIVE_WEB_MAX_SIZE]
                            # and map eventual RGB bitmaps on intensities
                            identifier += 1
            # okay this is about the intensity but what about the dimension scale axes
            # with tiff these could be stored in the tiff tag image metadata
            # print(np.shape(self.kikuchi))
            # print(self.kikuchi)
        return template

    def parse_pattern_stack_default_plot(self, template: dict) -> dict:
        """Parse data for the Kikuchi image stack default plot."""
        print("Parse Kikuchi pattern stack default plot...")
        trg = f"/ENTRY[entry{self.entry_id}]/simulation/IMAGE_SET_EM_KIKUCHI" \
              f"[image_set_em_kikuchi]/stack"

        template[f"{trg}/title"] = str("Kikuchi diffraction pattern stack")
        template[f"{trg}/@signal"] = "data_counts"
        template[f"{trg}/@axes"] = ["axis_x", "axis_y"]
        template[f"{trg}/@AXISNAME_indices[axis_x_indices]"] = np.uint32(0)
        template[f"{trg}/@AXISNAME_indices[axis_y_indices]"] = np.uint32(1)

        trg = f"/ENTRY[entry{self.entry_id}]/simulation/IMAGE_SET_EM_KIKUCHI" \
              f"[image_set_em_kikuchi]/stack/data_counts"
        template[f"{trg}"] = {"compress": self.stack, "strength": 1}
        # 0 is y while 1 is x !
        template[f"{trg}/@long_name"] = "Signal"
        template[f"{trg}/@CLASS"] = "IMAGE"  # required by H5Web to plot RGB maps
        template[f"{trg}/@IMAGE_VERSION"] = "1.2"
        template[f"{trg}/@SUBCLASS_VERSION"] = np.int64(15)

        axes_names = [("axis_x", 1, "x-axis"), ("axis_y", 0, "y-axis")]
        for axis in axes_names:
            trg = f"/ENTRY[entry{self.entry_id}]/simulation/IMAGE_SET_EM_KIKUCHI" \
                  f"[image_set_em_kikuchi]/stack/{axis[0]}"
            axis_i = np.asarray(
                np.linspace(0, self.stack_meta["size"][axis[1]],
                            num=self.stack_meta["size"][axis[1]],
                            endpoint=True), np.float64)
            # overwrite with calibrated scale if available
            # i.e. when self.stack_meta["axis_x"] not None:
            template[f"{trg}"] = {"compress": axis_i, "strength": 1}
            template[f"{trg}/@long_name"] = f"Pixel coordinate along {axis[2]}"
            template[f"{trg}/@units"] = "px"
        return template

    def parse(self, template: dict) -> dict:
        """Parse NOMAD OASIS relevant data and metadata from a ZIP file."""
        print("Parsing EBSD data generic-style for a ZIP example...")
        print(self.file_name)
        print(f"{self.entry_id}")
        self.parse_zip(template)
        self.parse_pattern_stack_default_plot(template)
        return template



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Utility functions for generation of EM NeXus datasets for dev purposes."""

# pylint: disable=no-member, R0801

import hashlib

import datetime

import math

import numpy as np

# import matplotlib.pyplot as plt

import hyperspy.api as hs

from ase.data import chemical_symbols

from pynxtools.dataconverter.readers.em_spctrscpy.utils.em_versioning \
    import NX_EM_ADEF_NAME, NX_EM_ADEF_VERSION, NX_EM_EXEC_NAME, NX_EM_EXEC_VERSION

# parameter affecting reconstructed positions and size
# default parameter
MAX_USERS = 4
# compose hypothetical spectrum (ignoring background)
XRAY_ENERGY_DELTA = 0.01  # in keV
PEAK_WIDTH_SIGMA = 0.1  # stddev in keV assumed the same for all peaks
SIGNAL_YIELD = 1000.  # arbitrarily assumed signal intensity


class EmSpctrscpyCreateExampleData:
    """A synthesized dataset meant to be used for development purposes only!."""

    def __init__(self, synthesis_id):
        # assure deterministic behaviour of the PRNG
        np.random.seed(seed=synthesis_id)

        self.n_entries = 1
        print("Generating one random example NXem entry...")
        self.entry_id = 1
        self.elements_observed = []
        self.e_axis = []
        self.cnts_summary = []
        # synthesizing realistic datasets for electron microscopy requires a
        # physical model of the image formation process see e.g. Kirkland et al.
        # for an (incomplete) view of this large topic
        # https://doi.org/10.1007/978-3-030-33260-0
        # we are interested in having spectra and maybe some synthesized images
        # as this is already sufficient to explore search capabilities in e.g. NOMAD Oasis

    def emulate_entry(self, template: dict) -> dict:
        """Copy data in entry section."""
        # print("Parsing entry...")
        trg = f"/ENTRY[entry{self.entry_id}]/"
        template[f"{trg}definition"] = NX_EM_ADEF_NAME
        template[f"{trg}@version"] = NX_EM_ADEF_VERSION
        template[f"{trg}PROGRAM[program1]/program"] = NX_EM_EXEC_NAME
        template[f"{trg}PROGRAM[program1]/program/@version"] = NX_EM_EXEC_VERSION
        template[f"{trg}start_time"] = datetime.datetime.now().astimezone().isoformat()
        template[f"{trg}end_time"] = datetime.datetime.now().astimezone().isoformat()

        msg = '''
              !!! WARNING: These are mocked data !!!
              These are meant to be used exclusively
              for verifying NOMAD search capabilities.
              '''
        template[f"{trg}experiment_description"] = msg
        template[f"{trg}experiment_documentation"] = "free text field"
        experiment_identifier = str(
            f"EM{np.random.choice(100, 1)[0]}/{np.random.choice(100000, 1)[0]}")
        template[f"{trg}experiment_identifier"] = experiment_identifier

        return template

    def emulate_user(self, template: dict) -> dict:
        """Copy data in user section."""
        # print("Parsing user...")
        prefix = f"/ENTRY[entry{self.entry_id}]/"
        user_names = np.unique(
            np.random.choice(["Sherjeel", "MarkusK", "Benedikt", "Johannes",
                              "Gerardo", "Kristiane", "Sabine", "Sophie", "Tom",
                              "Volker", "MarkusW", "PeterK", "Oonagh", "Annika",
                              "ChristophP", "Thomas", "Mariano", "Tilmann",
                              "ChristophF", "Niels", "Dieter", "Alexander",
                              "Katharina", "Florian", "Sebastian", "Sandor",
                              "Carola", "Chris", "Hampus", "Pepe", "Lauri",
                              "MarkusS", "Christoph", "Claudia"],
                             1 + np.random.choice(MAX_USERS, 1)))
        user_id = 1
        for name in user_names:
            trg = f"{prefix}USER[user{user_id}]/"
            template[f"{trg}name"] = str(name)
            user_id += 1

        return template

    def emulate_sample(self, template: dict) -> dict:
        """Copy data in specimen section."""
        # check if required fields exists and are valid
        # print("Parsing sample...")
        trg = f"/ENTRY[entry{self.entry_id}]/sample/"
        template[f"{trg}method"] = "simulation"

        self.elements_observed \
            = np.random.choice(np.asarray(np.linspace(1, 94, num=118, endpoint=True),
                                          np.uint32),
                               1 + int(np.random.uniform(low=0, high=5)))
        assert len(self.elements_observed) > 0, "List of assumed elements is empty!"
        unique_elements = set()
        for atomic_number in self.elements_observed:
            symbol = chemical_symbols[atomic_number]
            assert isinstance(symbol, str), "symbol is not a string!"
            if (symbol in chemical_symbols) & (symbol != "X"):
                unique_elements.add(symbol)
        print(f"Unique elements are: {list(unique_elements)}")
        template[f"{trg}atom_types"] = ", ".join(list(unique_elements))

        specimen_name = f"Mocked electron microscopy specimen " \
                        f"{np.random.choice(1000, 1)[0]}"
        template[f"{trg}name"] = specimen_name
        template[f"{trg}sample_history"] = "n/a"
        template[f"{trg}preparation_date"] \
            = datetime.datetime.now().astimezone().isoformat()
        template[f"{trg}short_title"] \
            = specimen_name.replace("Mocked atom probe specimen ", "")
        template[f"{trg}description"] = "n/a"

        template[f"{trg}thickness"] \
            = np.float64(np.max((np.random.normal(loc=40., scale=5.0), 10.)))
        template[f"{trg}thickness/@units"] = "nm"
        # template[f"{trg}density"] = 0.  # is optional
        # template[f"{trg}density/@units"] = "kg/m^3"

        return template

    def emulate_coordinate_system(self, template: dict) -> dict:
        """Define the coordinate systems to be used."""
        # print("Parsing coordinate system...")
        prefix = f"/ENTRY[entry{self.entry_id}]/" \
                 f"COORDINATE_SYSTEM_SET[coordinate_system_set]/"

        grpnm = f"{prefix}TRANSFORMATIONS[laboratory]/"
        cs_xyz = np.asarray(
            [[1., 0., 0.], [0., 1., 0.], [0., 0., 1.]], np.float64)
        cs_names = ["x", "y", "z"]
        for i in np.arange(0, 3):
            trg = f"{grpnm}AXISNAME[{cs_names[i]}]"
            template[trg] = cs_xyz[:, i]
            template[f"{trg}/@offset"] = np.asarray([0., 0., 0.], np.float64)
            template[f"{trg}/@offset_units"] = "m"
            template[f"{trg}/@depends_on"] = "."

        msg = '''
              This way of defining coordinate systems is an example
              how these can be defined. More discussion among members
              of FAIRmat Area A/B/C and members of the EM community
              plus more examples should be used to test how this
              feature of NeXus can be used.
              '''
        template[f"{prefix}@comment"] = msg

        return template

    def emulate_instrument_header(self, template: dict) -> dict:
        """Copy data in instrument header section."""
        # print("Parsing instrument header...")
        trg = f"/ENTRY[entry{self.entry_id}]/em_lab/"
        instrument_name = np.random.choice(
            ["Some ThermoFisher", "Some JEOL", "Some Zeiss",
             "Some TEscan", "Some Hitachi"], 1 + np.random.choice(1, 1))[0]
        template[f"{trg}instrument_name"] = str(instrument_name)
        template[f"{trg}location"] = str(np.random.choice(
            ["Berlin", "Leipzig", "Dresden", "Düsseldorf", "Aachen", "Garching",
             "Aachen", "Leoben", "Jülich"], 1 + np.random.choice(1, 1))[0])
        trg = f"/ENTRY[entry{self.entry_id}]/em_lab/FABRICATION[fabrication]/"
        template[f"{trg}vendor"] = instrument_name.replace("Some ", "")
        template[f"{trg}model"] = "n/a"
        template[f"{trg}identifier"] = str(hashlib.sha256(
            instrument_name.replace("Some ", "").encode("utf-8")).hexdigest())
        template[f"{trg}capabilities"] = "n/a"

        return template

    def emulate_ebeam_column(self, template: dict) -> dict:
        """Copy data in ebeam_column section."""
        # print("Parsing ebeam column...")
        prefix = f"/ENTRY[entry{self.entry_id}]/em_lab/EBEAM_COLUMN[ebeam_column]/"
        trg = f"{prefix}electron_source/"

        template[f"{trg}voltage"] = np.float64(
            np.random.choice(np.linspace(10., 300., num=30, endpoint=True), 1)[0])
        template[f"{trg}voltage/@units"] = "kV"
        template[f"{trg}emitter_type"] \
            = str(np.random.choice(["thermionic", "schottky", "field_emission"], 1)[0])

        # for aperture in apertures:
        trg = f"{prefix}APERTURE_EM[aperture_em1]/"
        template[f"{trg}value"] = np.uint32(
            np.random.choice(np.linspace(1, 5, num=5, endpoint=True), 1)[0])
        template[f"{trg}name"] = "aperture1"
        template[f"{trg}description"] = "n/a"

        # the above-mentioned snippet is a blue-print for lenses also...

        # corrector
        trg = f"{prefix}aberration_correction/"
        template[f"{trg}applied"] = bool(np.random.choice([0, 1], 1)[0])

        return template

    def emulate_ibeam_column(self, template: dict) -> dict:
        """Copy data in ibeam_column section."""
        # print("Parsing ibeam column...")
        return template

    def emulate_ebeam_deflector(self, template: dict) -> dict:
        """Copy data in ebeam_deflector section."""
        # print("Parsing ebeam deflector...")
        return template

    def emulate_ibeam_deflector(self, template: dict) -> dict:
        """Copy data in ibeam_deflector section."""
        # print("Parsing ibeam deflector...")
        return template

    def emulate_optics(self, template: dict) -> dict:
        """Copy data in optical_system_em section."""
        # print("Parsing optics...")
        trg = f"/ENTRY[entry{self.entry_id}]/em_lab/OPTICAL_SYSTEM_EM[optical_system_em]/"
        template[f"{trg}beam_current_description"] = "undefined"
        template[f"{trg}camera_length"] \
            = np.float64(np.random.normal(loc=1.0, scale=0.05))
        template[f"{trg}camera_length/@units"] = "m"
        template[f"{trg}magnification"] \
            = np.float64(np.random.choice([100., 1000., 10000., 100000.], 1)[0])
        template[f"{trg}defocus"] = np.float64(np.random.normal(loc=1.0e-7, scale=0.1e-7))
        template[f"{trg}defocus/@units"] = "m"
        template[f"{trg}semi_convergence_angle"] = np.float64(
            np.min((np.random.normal(loc=10., scale=1.), 1.0)))
        template[f"{trg}semi_convergence_angle/@units"] = "degree"
        template[f"{trg}working_distance"] = np.float64(
            np.trunc(np.random.choice(np.linspace(5., 20., num=15, endpoint=True), 1)[0]))
        template[f"{trg}working_distance/@units"] = "cm"
        template[f"{trg}beam_current"] = np.float64(
            np.min((np.random.normal(loc=10., scale=2.), 1.0)))
        template[f"{trg}beam_current/@units"] = "pA"

        return template

    def emulate_detector(self, template: dict) -> dict:
        """Copy data in detector section."""
        # print("Parsing detector...")
        detectors = np.unique(np.random.choice(
            ["SE", "BSE", "EBSD", "EDX", "INLINE"],
            1 + np.random.choice(5, 1)))
        detector_id = 1
        for detector in detectors:
            trg = f"/ENTRY[entry{self.entry_id}]/em_lab/DETECTOR[detector{detector_id}]/"
            template[f"{trg}local_name"] = str(detector)
            detector_id += 1

        return template

    def emulate_stage_lab(self, template: dict) -> dict:
        """Copy data in stage lab section."""
        # print("Parsing stage lab...")
        trg = f"/ENTRY[entry{self.entry_id}]/em_lab/stage_lab/"
        stage_name = np.random.choice(
            ["side_entry", "top_entry", "single_tilt", "quick_change",
             "multiple_specimen", "bulk_specimen", "double_tilt", "tilt_rotate",
             "heating_chip", "atmosphere_chip", "electrical_biasing_chip",
             "liquid_cell_chip"], 1)[0]
        template[f"{trg}name"] = str(stage_name)
        return template

    def emulate_random_input_from_eln(self, template: dict) -> dict:
        """Emulate random input as could come from an ELN."""
        print("Parsing random input from ELN...")
        self.emulate_entry(template)
        self.emulate_user(template)
        self.emulate_sample(template)
        self.emulate_coordinate_system(template)
        self.emulate_instrument_header(template)
        self.emulate_ebeam_column(template)
        self.emulate_ibeam_column(template)
        self.emulate_ebeam_deflector(template)
        self.emulate_ibeam_deflector(template)
        self.emulate_optics(template)
        self.emulate_detector(template)
        self.emulate_stage_lab(template)

        return template

    def emulate_random_xray_spectrum(self, template: dict) -> dict:
        """Emulate one event data group for Xray spectroscopy."""
        # !! data meant exclusively to be used for verification purposes !!
        assert len(self.elements_observed > 0), "No elements were observed !"
        composition = np.random.uniform(
            low=0., high=1., size=(len(self.elements_observed),))
        composition = composition / np.sum(composition)
        signal_contributions = []
        idx = 0
        # sample hypothetically X-ray lines for each element observed
        for atomic_number in self.elements_observed:
            symbol = chemical_symbols[atomic_number]
            if symbol in hs.material.elements:
                if 'Atomic_properties' in hs.material.elements[symbol]:
                    if 'Xray_lines' in hs.material.elements[symbol].Atomic_properties:
                        lines = hs.material.elements[symbol].Atomic_properties.Xray_lines
                        for xline_name, xline_props in lines.as_dictionary().items():
                            # print(key + ", " + str(value["weight"]))
                            signal_contributions.append((atomic_number,
                                                         symbol,
                                                         composition[idx],
                                                         xline_name,
                                                         xline_props["energy (keV)"],
                                                         xline_props["weight"]))
            idx += 1
        # self.elements_observed = np.unique(symbols)

        xray_energy_max = 0.
        for tpl in signal_contributions:
            xray_energy_max = np.max((xray_energy_max, tpl[4]))
        # print(xray_energy_max)

        n_bins = int(np.ceil(
            (xray_energy_max + 3. * XRAY_ENERGY_DELTA + 1.)
            / XRAY_ENERGY_DELTA))  # covering [0., n_bins * XRAY_ENERGY_DELTA]
        self.e_axis = np.linspace(
            0.5 * XRAY_ENERGY_DELTA,
            0.5 * XRAY_ENERGY_DELTA + n_bins * XRAY_ENERGY_DELTA,
            num=n_bins, endpoint=True)

        self.cnts_summary = np.zeros((n_bins,), np.float64)
        for tpl in signal_contributions:
            # idx = np.abs(self.e_axis - tpl[4]).argmin()
            # integrate analytically, assume Gaussian peak with stddev PEAK_WIDTH_SIGMA
            cnts_tpl = np.zeros((n_bins, ), np.float64)
            for idx in np.arange(0, n_bins):
                cnts_tpl[idx] = SIGNAL_YIELD * tpl[2] * tpl[5] * 0.5 \
                    * (math.erf(1. / (np.sqrt(2.) * PEAK_WIDTH_SIGMA)
                       * (tpl[4] - (0. + idx * XRAY_ENERGY_DELTA)))
                       - math.erf(1. / (np.sqrt(2.) * PEAK_WIDTH_SIGMA)
                       * (tpl[4] - (XRAY_ENERGY_DELTA + idx * XRAY_ENERGY_DELTA))))
            self.cnts_summary = np.add(self.cnts_summary, cnts_tpl)
        # plt.plot(self.e_axis, self.cnts_summary)
        # plt.xlabel("energy (keV)")
        # plt.ylabel("cnts")
        # plt.xscale("log")

        trg = f"/ENTRY[entry{self.entry_id}]/measurement/" \
              f"EVENT_DATA_EM[event_data_em1]/xray/summary/"
        template[f"{trg}title"] = "Accumulated X-ray spectrum"
        # template[f"{trg}@long_name"] = "Xray"
        template[f"{trg}@signal"] = "data_counts"
        template[f"{trg}@axes"] = ["axis_photon_energy"]
        template[f"{trg}@AXISNAME_indices[axis_photon_energy_indices]"] = 0
        template[f"{trg}DATA[data_counts]"] \
            = {"compress": self.cnts_summary, "strength": 1}
        template[f"{trg}DATA[data_counts]/@units"] = ""
        template[f"{trg}DATA[data_counts]/@long_name"] = "Photon counts (1)"
        template[f"{trg}AXISNAME[axis_photon_energy]"] \
            = {"compress": self.e_axis, "strength": 1}
        template[f"{trg}AXISNAME[axis_photon_energy]/@units"] = "keV"
        template[f"{trg}AXISNAME[axis_photon_energy]/@long_name"] = "Photon energy (keV)"
        return template

    def synthesize(self, template: dict) -> dict:
        """Hand-over instantiated dataset to dataconverter template."""
        for entry_id in np.arange(1, self.n_entries + 1):
            self.entry_id = entry_id
            print(f"Generating entry {self.entry_id}...")

            # metadata
            self.emulate_random_input_from_eln(template)

            # heavy numerical data, here the synthesized "measurement" data
            self.emulate_random_xray_spectrum(template)

        return template



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Parser for loading generic orientation microscopy data based on ."""

# pylint: disable=no-member

from typing import Tuple, Any

from pynxtools.dataconverter.readers.base.reader import BaseReader

from pynxtools.dataconverter.readers.em_nion.utils.swift_define_io_cases \
    import EmNionUseCaseSelector

from pynxtools.dataconverter.readers.em_nion.utils.swift_load_generic_eln \
    import NxEmNionElnSchemaParser

from pynxtools.dataconverter.readers.em_nion.utils.swift_zipped_project_parser \
    import NxEmNionSwiftProjectParser

from pynxtools.dataconverter.readers.em_spctrscpy.utils.em_nexus_plots \
    import em_spctrscpy_default_plot_generator


class EmNionReader(BaseReader):
    """Parse content from nionswift projects.

    Specifically, electron microscopy
    towards a NXem.nxdl-compliant NeXus file.
    """

    # pylint: disable=too-few-public-methods

    # Whitelist for the NXDLs that the reader supports and can process
    supported_nxdls = ["NXem"]

    # pylint: disable=duplicate-code
    def read(self,
             template: dict = None,
             file_paths: Tuple[str] = None,
             objects: Tuple[Any] = None) -> dict:
        """Read data from given file, return filled template dictionary em."""
        # pylint: disable=duplicate-code
        template.clear()

        # this em_om parser combines multiple sub-parsers
        # so we need the following input:
        # logical analysis which use case
        # data input from an ELN (using an ELN-agnostic) YAML representation
        # data input from technology partner files, here zipped nionswift project
        # directory and file renamed from ending with zip to nszip
        # functionalities for creating default plots

        entry_id = 1

        print("Parse ELN and compressed nionswift project content...")
        case = EmNionUseCaseSelector(file_paths)
        if case.is_valid is False:
            print("Such a combination of input-file(s, if any) is not supported !")
            return {}

        print("Parse (meta)data coming from an ELN...")
        if case.eln_parser_type == "generic":
            eln = NxEmNionElnSchemaParser(case.eln[0], entry_id)
            eln.parse(template)
        else:
            print("No interpretable ELN input found!")
            return {}

        print("Parse (numerical) data and metadata from nionswift project...")
        if case.prj_parser_type == "nionswift":
            swift_parser = NxEmNionSwiftProjectParser(case.prj[0], entry_id)
            swift_parser.parse(template)
        else:
            print("No input-file defined for technology partner data !")

        # at this point the default plots exist already
        # we only need to decorate the template to point to the mandatory ROI overview
        print("Create NeXus default plottable data...")
        em_spctrscpy_default_plot_generator(template, entry_id)

        debugging = False
        if debugging is True:
            print("Reporting state of template before passing to HDF5 writing...")
            for keyword in template.keys():
                print(keyword)
                # print(type(template[keyword]))
                # print(f"{keyword}, {template[keyword]}")

        print("Forward instantiated template to the NXS writer...")
        return template


# This has to be set to allow the convert script to use this reader.
READER = EmNionReader



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Conventions suggested by D. Rowenhorst et al. in DOI: 10.1088/0965-0393/23/8/083501."""

# pylint: disable=no-member,duplicate-code

msmse_convention = {
    "three_dimensional_rotation_handedness": "counter_clockwise",
    "rotation_convention": "passive",
    "euler_angle_convention": "zxz",
    "axis_angle_convention": "rotation_angle_on_interval_zero_to_pi"
}
# the sign convention is mentioned in the paper but left as a parameter
# "sign_convention": "p_minus_one"


def is_consistent_with_msmse_convention(dct):
    """Checks if a set of conventions is consistent with that paper (see above)."""
    # triple boolean, yes, no, unclear
    req_fields = [
        "three_dimensional_rotation_handedness",
        "rotation_convention",
        "euler_angle_convention",
        "axis_angle_convention"]
    for field_name in req_fields:
        if (field_name not in dct) or (field_name not in msmse_convention):
            return "unclear"
    # okay, so either matching or not
    for field_name in req_fields:
        if dct[field_name] == msmse_convention[field_name]:
            continue
        return "inconsistent"
    return "consistent"



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Parse references from rii yaml files"""
from dataclasses import dataclass, field
import re
from typing import Any, Dict, Optional
import requests_cache


@dataclass
class Citation:
    """A representation for a citation"""

    _raw_str: str = field(repr=False)
    ref_str: str
    url: Optional[str]
    doi: Optional[str]
    session: Optional[requests_cache.CachedSession] = None
    bibtex: Optional[str] = None

    @staticmethod
    def parse_citations(reference: str, download_bibtex: bool = False):
        """Parses a reference string into its components"""
        cites = re.split(r"<br\s*\/?\s*>", reference)

        clean_cites = []
        for cite in cites:
            clean_cites.append(Citation(cite, download_bibtex))
        return clean_cites

    def __init__(
        self, ref_str: str, download_bibtex: bool = False, desc: Optional[str] = None
    ):
        self._raw_str = ref_str
        self.ref_str = re.sub(
            r"\<\/?[^\<\>]*\>",
            "",
            re.search(r"(?:\d\))?\s*(.*)", ref_str).group(1),
        )
        self.get_non_doi_url()
        self.get_doi()

        if desc:
            self.description = desc
        else:
            self.description = (
                "Literature reference from which this "
                "dispersion function was extracted."
            )

        if download_bibtex:
            self.download_bibtex()

    def download_bibtex(self):
        """Requests the bibtext entry for a given doi"""
        if not self.doi:
            return

        if self.session is None:
            self.session = requests_cache.CachedSession("rii_bibtex")

        req = self.session.get(
            f"https://doi.org/{self.doi}",
            timeout=30,
            headers={"Accept": "application/x-bibtex"},
        )
        if req.status_code == 200:
            self.bibtex = req.text

    def get_non_doi_url(self):
        """Extract urls which are not dois"""
        url = re.search(r"href\=\"([^\"]+)\"", self._raw_str)
        self.url = url.group(1) if url and "doi.org" not in url.group(1) else None

    def get_doi(self):
        """Extracts doi from a reference string"""
        doi_match = re.search(r"\".*doi\.org\/([^\"]+)\"", self._raw_str)
        self.doi = doi_match.group(1) if doi_match else None

    def write_entries(self, template: Dict[str, Any], path: str):
        """Write entries to pynxtools template"""
        reference_identifier = "REFERENCES"
        name = basename = "reference"

        while f"{path}/{reference_identifier}[{name}]/text" in template:
            count = locals().get("count", 0) + 1
            name = f"{basename}{count}"

        base_path = f"{path}/{reference_identifier}[{name}]"
        if self.url:
            template[f"{base_path}/url"] = self.url
        if self.doi:
            template[f"{base_path}/doi"] = self.doi
        if self.bibtex:
            template[f"{base_path}/bibtex"] = self.bibtex

        template[f"{base_path}/text"] = self.ref_str
        template[f"{base_path}/description"] = self.description



import shutil
from pathlib import Path
from typing import List

from ..docs import AnchorRegistry
from ..docs import NXClassDocGenerator
from ..globals import directories
from ..nxdl import find_definition
from ..nxdl import validate_definition
from ..utils.diff import diff_ascii


def nxclass_args(parser):
    parser.add_argument(
        "name",
        type=str,
        help="NeXus class name (For example 'nxdata')",
    )
    parser.add_argument(
        "--test",
        action="store_true",
        help="Validate the NeXus class definition",
    )
    parser.add_argument(
        "--prepare",
        action="store_true",
        help="Save the NeXus class documentation in the build directory",
    )
    parser.add_argument(
        "--diff",
        action="store_true",
        help="Print changes in the NeXus class documentation",
    )
    parser.add_argument(
        "--print",
        action="store_true",
        help="Print the NeXus class documentation",
    )


def nxclass_exec(args):
    nxdl_file = find_definition(args.name)
    assert nxdl_file, f"No definition found for {args.name}"

    if args.test:
        validate_definition(nxdl_file)

    if args.prepare or args.diff or args.print:
        generator = NXClassDocGenerator()
        if args.prepare:
            output_path = directories.manual_build_staticroot()
        else:
            output_path = None
        anchor_registry = AnchorRegistry(output_path=output_path)
        rst_lines = generator(nxdl_file, anchor_registry=anchor_registry)

    if args.prepare:
        save_nxclass_docs(nxdl_file, rst_lines)
        print("add to anchor list files in", output_path)
        anchor_registry.write()

    if args.diff:
        diff_nxclass_docs(nxdl_file, rst_lines)

    if args.print:
        print("".join(rst_lines))


def save_nxclass_docs(nxdl_file: Path, rst_lines: List[str]) -> None:
    """Build the NXDL file: this means prepare the documentation
    and resources in the build directory.
    """
    rst_file_name = directories.get_rst_filename(nxdl_file)

    # Save the documentation in the build directory
    print("generate", nxdl_file)
    print("      ->", rst_file_name)
    rst_file_name.parent.mkdir(parents=True, exist_ok=True)
    with open(rst_file_name, "w") as fh:
        fh.writelines(rst_lines)

    # Copy resources to the build directory
    resource_name = rst_file_name.stem[2:]  # (e.g. NXaperture -> aperture)
    nxdl_resource_dir = nxdl_file.parent / resource_name
    if nxdl_resource_dir.is_dir():
        rst_resource_dir = rst_file_name.parent / resource_name
        print("copy", nxdl_resource_dir)
        print("  ->", rst_resource_dir)
        shutil.copytree(nxdl_resource_dir, rst_resource_dir, dirs_exist_ok=True)

    # Copy the NXDL file as it might be used in a "literalinclude"
    nxdl_file2 = rst_file_name.parent / nxdl_file.name
    print("generate", nxdl_file)
    print("      ->", nxdl_file2)
    shutil.copyfile(nxdl_file, nxdl_file2)


def diff_nxclass_docs(nxdl_file: Path, rst_lines: List[str]) -> None:
    """Build the NXDL file: this means prepare the documentation
    and resources in the build directory.
    """
    rst_file_name = directories.get_rst_filename(nxdl_file)
    diff_ascii(nxdl_file, rst_lines, rst_file_name)



import numpy as np
from nexusformat.nexus import *


def write_nexus_file(fname, image, md={}):
    """
    write the image to a NeXus HDF5 data file

    Parameters
    ----------
    fname : str
        name of the file (relative or absolute) to be written
    image : numpy array
        the image data
    md : dictionary
        key: value where value is something that can be written by h5py
             (such as str, int, float, numpy array, ...)
    """
    nx = NXroot()
    nx['/entry'] = NXentry(NXinstrument(NXdetector()))
    nx['entry/instrument/detector/image'] = NXfield(image, units='counts',
                                                    compression='gzip')
    nx['entry/data'] = NXdata()
    nx['entry/data'].makelink(nx['entry/instrument/detector/image'])
    nx['entry/data'].nxsignal = nx['entry/data/image']

    if len(md) > 0:
        # /entry/instrument/metadata (optional, for metadata)
        metadata = nx['/entry/instrument/metadata'] = NXcollection()
        for k, v in md.items():
            metadata[k] = v

    nx.save(fname, 'w')

	
if __name__ == "__main__":
	"""demonstrate how to use this code"""
	import epics
	prefix = "13SIM1:"
	img = epics.caget(prefix+"image1:ArrayData")
	size_x = epics.caget(prefix+"cam1:ArraySizeX_RBV")
	size_y = epics.caget(prefix+"cam1:ArraySizeY_RBV")
	# edit the full image for just the binned data
	img = img[:size_x*size_y].reshape((size_x, size_y))

	extra_information = dict(
		unique_id = epics.caget(prefix+"image1:UniqueId_RBV"),
		size_x = size_x,
		size_y = size_y,
		detector_state = epics.caget(prefix+"cam1:DetectorState_RBV"),
		bitcoin_value="15000",
	)
	write_nexus_file("example.h5", img, md=extra_information)



"""Reads NeXus HDF5 files using h5py and prints the contents"""

import h5py  # HDF5 support


testFiles = ("simple_example_test.nexus.hdf5",)

GROUP_TYPE_MATCH = "<class 'h5py.highlevel.Group'>"
DATASET_TYPE_MATCH = "<class 'h5py.highlevel.Dataset'>"


def print_attr(parent, label):
    for k, v in parent.attrs.items():
        print("%s.attrs[%s]=%s" % (label, k, v))


def print_child(item, label):
    print("#" + "-" * 40)
    print("%s  %s" % (label, type(item)))
    print_attr(item, label)
    if repr(type(item)) == GROUP_TYPE_MATCH:
        base = label
        if "NX_class" in item.attrs:
            base += ":" + item.attrs["NX_class"]
        for k in item.keys():
            key = "%s/%s" % (base, k)
            print_child(item[k], key)
    if repr(type(item)) == DATASET_TYPE_MATCH:
        # print label, item.value
        print(label)
        print("shape:", item.shape)
        print("size:", len(item.shape))
        print("NumPy dtype:", item.dtype)


def process(fileName):
    try:
        f = h5py.File(fileName, "r")
    except Exception:
        return False
    try:
        print(f.filename)
        print("keys: ", f.keys())
        print_attr(f, "f")
        for k in f.keys():
            print_child(f[k], "%s://%s" % (fileName, k))
    finally:
        f.close()
    return True


if __name__ == "__main__":
    for fileName in testFiles:
        print("#" + "=" * 60)
        if not process(fileName):
            print("Could not open:", fileName)



#!/usr/bin/env python3
"""This tool accomplishes some tests for the yaml2nxdl parser

"""
#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import os
import sys
import filecmp
from datetime import datetime
from pathlib import Path
import xml.etree.ElementTree as ET
import pytest
from click.testing import CliRunner
import pynxtools.nyaml2nxdl.nyaml2nxdl as nyml2nxdl
from pynxtools.nyaml2nxdl import nyaml2nxdl_forward_tools


def delete_duplicates(list_of_matching_string):
    """
    Delete duplicate from lists
    """
    return list(dict.fromkeys(list_of_matching_string))


def check_file_fresh_baked(test_file):
    """
    Get sure that the test file is generated by the converter
    """
    path = Path(test_file)
    timestamp = datetime.fromtimestamp(path.stat().st_mtime).strftime("%d/%m/%Y %H:%M")
    now = datetime.now().strftime("%d/%m/%Y %H:%M")
    assert timestamp == now, 'xml file not generated'


def find_matches(xml_file, desired_matches):
    """
        Read xml file and find desired matches. Return a list of two lists in the form:
    [[matching_line],[matching_line_index]]
    """
    with open(xml_file, 'r') as file:
        xml_reference = file.readlines()
    lines = []
    lines_index = []
    found_matches = []
    for i, line in enumerate(xml_reference):
        for desired_match in desired_matches:
            if str(desired_match) in str(line):
                lines.append(line)
                lines_index.append(i)
                found_matches.append(desired_match)
    # ascertain that all the desired matches were found in file
    found_matches_clean = delete_duplicates(found_matches)
    assert len(found_matches_clean) == len(desired_matches), 'some desired_matches were \
not found in file'
    return [lines, lines_index]


def compare_matches(ref_xml_file, test_yml_file, test_xml_file, desired_matches):
    """
        Check if a new xml file is generated
    and if test xml file is equal to reference xml file
    """
    # Reference file is read
    ref_matches = find_matches(ref_xml_file, desired_matches)
    # Test file is generated
    runner = CliRunner()
    result = runner.invoke(nyml2nxdl.launch_tool, ['--input-file', test_yml_file])
    assert result.exit_code == 0
    check_file_fresh_baked(test_xml_file)
    # Test file is read
    test_matches = find_matches(test_xml_file, desired_matches)
    assert test_matches == ref_matches


def test_links():
    """
    Check the correct parsing of links
    """
    data_path = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                             '../data/nyaml2nxdl')
    ref_xml_link_file = 'tests/data/nyaml2nxdl/Ref_NXtest_links.nxdl.xml'
    test_yml_link_file = 'tests/data/nyaml2nxdl/NXtest_links.yaml'
    test_xml_link_file = 'tests/data/nyaml2nxdl/NXtest_links.nxdl.xml'
    # ref_xml_link_file = os.path.abspath(data_path + '/Ref_NXtest_links.nxdl.xml')
    # test_yml_link_file = os.path.abspath(data_path + '/NXtest_links.yaml')
    # test_xml_link_file = os.path.abspath(data_path + '/NXtest_links.nxdl.xml')
    desired_matches = ['<link', '/>']
    compare_matches(
        ref_xml_link_file,
        test_yml_link_file,
        test_xml_link_file,
        desired_matches)
    os.remove('tests/data/nyaml2nxdl/NXtest_links.nxdl.xml')
    sys.stdout.write('Test on links okay.\n')


def test_docs():
    """In this test an xml file in converted to yml and then back to xml.
    The xml trees of the two files are then compared.
    """
    ref_xml_file = 'tests/data/nyaml2nxdl/Ref_NXellipsometry-docCheck.nxdl.xml'
    test_yml_file = 'tests/data/nyaml2nxdl/NXellipsometry-docCheck.yaml'
    test_xml_file = 'tests/data/nyaml2nxdl/NXellipsometry-docCheck.nxdl.xml'
    desired_matches = ['<doc', '</doc>']
    compare_matches(
        ref_xml_file,
        test_yml_file,
        test_xml_file,
        desired_matches)
    os.remove('tests/data/nyaml2nxdl/NXellipsometry-docCheck.nxdl.xml')
    sys.stdout.write('Test on documentation formatting okay.\n')


def test_nxdl2yaml_doc_format_and_nxdl_part_as_comment():
    """
    This test for two reason:
        1. In test-1 an nxdl file with all kind of doc formats are translated
    to yaml to check if they are correct.
        2. In test-2: Check the nxdl that comes at the end of yaml file as comment.
    """
    ref_xml_file = 'tests/data/nyaml2nxdl/Ref_NXentry.nxdl.xml'
    ref_yml_file = 'tests/data/nyaml2nxdl/Ref_NXentry.yaml'
    test_yml_file = 'tests/data/nyaml2nxdl/Ref_NXentry_parsed.yaml'
    result = CliRunner().invoke(nyml2nxdl.launch_tool, ['--input-file', ref_xml_file])
    assert result.exit_code == 0
    check_file_fresh_baked(test_yml_file)

    result = filecmp.cmp(ref_yml_file, test_yml_file, shallow=False)
    assert result, 'Ref YML and parsed YML\
has not the same structure!!'
    os.remove(test_yml_file)
    sys.stdout.write('Test on xml -> yml doc formatting okay.\n')


def test_fileline_error():
    """
    In this test the yaml fileline in the error message is tested.
    """
    test_yml_file = 'tests/data/nyaml2nxdl/NXfilelineError1.yaml'
    out_nxdl = 'tests/data/nyaml2nxdl/NXfilelineError1.nxdl.xml'
    out_yaml = 'tests/data/nyaml2nxdl/temp_NXfilelineError1.yaml'
    result = CliRunner().invoke(nyml2nxdl.launch_tool, ['--input-file', test_yml_file])
    assert result.exit_code == 1
    assert '13' in str(result.exception)
    os.remove(out_nxdl)
    os.remove(out_yaml)

    test_yml_file = 'tests/data/nyaml2nxdl/NXfilelineError2.yaml'
    out_nxdl = 'tests/data/nyaml2nxdl/NXfilelineError2.nxdl.xml'
    out_yaml = 'tests/data/nyaml2nxdl/temp_NXfilelineError2.yaml'
    result = CliRunner().invoke(nyml2nxdl.launch_tool, ['--input-file', test_yml_file])
    assert result.exit_code == 1
    assert '21' in str(result.exception)
    os.remove(out_nxdl)
    os.remove(out_yaml)

    test_yml_file = 'tests/data/nyaml2nxdl/NXfilelineError3.yaml'
    out_nxdl = 'tests/data/nyaml2nxdl/NXfilelineError3.nxdl.xml'
    out_yaml = 'tests/data/nyaml2nxdl/temp_NXfilelineError3.yaml'
    result = CliRunner().invoke(nyml2nxdl.launch_tool, ['--input-file', test_yml_file])
    assert result.exit_code == 1
    assert '25' in str(result.exception)
    os.remove(out_nxdl)
    os.remove(out_yaml)

    sys.stdout.write('Test on xml -> yml fileline error handling okay.\n')


def test_symbols():
    """
    Check the correct parsing of symbols
    """
    ref_xml_symbol_file = 'tests/data/nyaml2nxdl/Ref_NXnested_symbols.nxdl.xml'
    test_yml_symbol_file = 'tests/data/nyaml2nxdl/NXnested_symbols.yaml'
    test_xml_symbol_file = 'tests/data/nyaml2nxdl/NXnested_symbols.nxdl.xml'
    desired_matches = ['<symbols>', '</symbols>', '<symbols']
    compare_matches(
        ref_xml_symbol_file,
        test_yml_symbol_file,
        test_xml_symbol_file,
        desired_matches)
    os.remove('tests/data/nyaml2nxdl/NXnested_symbols.nxdl.xml')
    sys.stdout.write('Test on symbols okay.\n')


def test_attributes():
    """
        Check expected attributes in NeXus fields, groups, and attributes.
        Check proper doc elements.
    """
    ref_xml_attribute_file = 'tests/data/nyaml2nxdl/Ref_NXattributes.nxdl.xml'
    test_yml_attribute_file = 'tests/data/nyaml2nxdl/NXattributes.yaml'
    test_xml_attribute_file = 'tests/data/nyaml2nxdl/NXattributes.nxdl.xml'
    desired_matches = ['<attribute', '</attribute>', '<doc>', '</doc>',
                       '<field', '</field>', '<group', '</group>']
    compare_matches(
        ref_xml_attribute_file,
        test_yml_attribute_file,
        test_xml_attribute_file,
        desired_matches)
    os.remove('tests/data/nyaml2nxdl/NXattributes.nxdl.xml')
    sys.stdout.write('Test on attributes okay.\n')


def test_extends():
    """
        Check the correct handling of extends keyword
    """
    ref_xml_attribute_file = 'tests/data/nyaml2nxdl/Ref_NXattributes.nxdl.xml'
    test_yml_attribute_file = 'tests/data/nyaml2nxdl/NXattributes.yaml'
    test_xml_attribute_file = 'tests/data/nyaml2nxdl/NXattributes.nxdl.xml'
    runner = CliRunner()
    result = runner.invoke(nyml2nxdl.launch_tool, ['--input-file', test_yml_attribute_file])
    assert result.exit_code == 0
    ref_root_node = ET.parse(ref_xml_attribute_file).getroot()
    test_root_node = ET.parse(test_xml_attribute_file).getroot()
    assert ref_root_node.attrib == test_root_node.attrib
    os.remove('tests/data/nyaml2nxdl/NXattributes.nxdl.xml')
    sys.stdout.write('Test on extends keyword okay.\n')


def test_symbols_and_enum_docs():
    """
        Check the correct handling of empty attributes
    or attributes fields, e.g. doc
    """
    ref_xml_file = 'tests/data/nyaml2nxdl/Ref_NXmytests.nxdl.xml'
    test_yml_file = 'tests/data/nyaml2nxdl/NXmytests.yaml'
    test_xml_file = 'tests/data/nyaml2nxdl/NXmytests.nxdl.xml'
    desired_matches = ['<attribute', '</attribute>', '<doc>', '</doc>',
                       '<symbols>', '</symbols>', '<symbols',
                       '<dimensions', '</dimensions>', '<dim']
    compare_matches(
        ref_xml_file,
        test_yml_file,
        test_xml_file,
        desired_matches)
    os.remove('tests/data/nyaml2nxdl/NXmytests.nxdl.xml')
    sys.stdout.write('Test on docs in enumeration and symbols okay.\n')


def test_xml_parsing():
    """
        In this test an xml file in converted to yml and then back to xml.
    The xml trees of the two files are then compared.
    """
    ref_xml_file = 'tests/data/nyaml2nxdl/Ref_NXellips.nxdl.xml'
    test_yml_file = 'tests/data/nyaml2nxdl/Ref_NXellips_parsed.yaml'
    test_xml_file = 'tests/data/nyaml2nxdl/Ref_NXellips_parsed.nxdl.xml'
    result = CliRunner().invoke(nyml2nxdl.launch_tool, ['--input-file', ref_xml_file])
    assert result.exit_code == 0
    check_file_fresh_baked(test_yml_file)
    result = CliRunner().invoke(nyml2nxdl.launch_tool, ['--input-file', test_yml_file])
    assert result.exit_code == 0
    check_file_fresh_baked(test_xml_file)

    test_tree = ET.parse(test_xml_file)
    test_tree_flattened = set([i.tag.split("}", 1)[1] for i in test_tree.iter()])

    ref_tree = ET.parse(ref_xml_file)
    ref_tree_flattened = set([i.tag.split("}", 1)[1] for i in ref_tree.iter()])

    assert test_tree_flattened == ref_tree_flattened, 'Ref XML and parsed XML\
has not the same tree structure!!'
    os.remove('tests/data/nyaml2nxdl/Ref_NXellips_parsed.nxdl.xml')
    os.remove('tests/data/nyaml2nxdl/Ref_NXellips_parsed.yaml')
    sys.stdout.write('Test on xml -> yml -> xml okay.\n')


def test_yml_parsing():
    """In this test an xml file in converted to yml and then back to xml.
    The xml trees of the two files are then compared.
    """
    ref_yml_file = 'tests/data/nyaml2nxdl/Ref_NXellipsometry.yaml'
    test_xml_file = 'tests/data/nyaml2nxdl/Ref_NXellipsometry.nxdl.xml'
    test_yml_file = 'tests/data/nyaml2nxdl/Ref_NXellipsometry_parsed.yaml'
    result = CliRunner().invoke(nyml2nxdl.launch_tool, ['--input-file', ref_yml_file])
    assert result.exit_code == 0
    check_file_fresh_baked(test_xml_file)
    result = CliRunner().invoke(nyml2nxdl.launch_tool, ['--input-file', test_xml_file])
    assert result.exit_code == 0
    check_file_fresh_baked(test_yml_file)

    test_yml_tree = nyaml2nxdl_forward_tools.yml_reader(test_yml_file)

    ref_yml_tree = nyaml2nxdl_forward_tools.yml_reader(ref_yml_file)

    assert list(test_yml_tree) == list(ref_yml_tree), 'Ref YML and parsed YML \
has not the same root entries!!'
    os.remove('tests/data/nyaml2nxdl/Ref_NXellipsometry_parsed.yaml')
    os.remove('tests/data/nyaml2nxdl/Ref_NXellipsometry.nxdl.xml')
    sys.stdout.write('Test on yml -> xml -> yml okay.\n')


def test_yml_consistency_comment_parsing():
    """Test comments parsing from yaml. Convert 'yaml' input file to '.nxdl.xml' and
    '.nxdl.xml' to '.yaml'
    """
    from pynxtools.nyaml2nxdl.comment_collector import CommentCollector
    from pynxtools.nyaml2nxdl.nyaml2nxdl_helper import LineLoader

    ref_yml_file = 'tests/data/nyaml2nxdl/Ref_NXcomment.yaml'
    test_yml_file = 'tests/data/nyaml2nxdl/Ref_NXcomment_consistency.yaml'

    result = CliRunner().invoke(nyml2nxdl.launch_tool,
                                ['--input-file', ref_yml_file,
                                 '--check-consistency'])
    assert result.exit_code == 0, (f'Exception: {result.exception}, \nExecution Info:'
                                   '{result.exc_info}')
    with open(ref_yml_file, 'r', encoding='utf-8') as ref_yml:
        loader = LineLoader(ref_yml)
        ref_loaded_yaml = loader.get_single_data()
    ref_comment_blocks = CommentCollector(ref_yml_file, ref_loaded_yaml)
    ref_comment_blocks.extract_all_comment_blocks()

    with open(test_yml_file, 'r', encoding='utf-8') as test_yml:
        loader = LineLoader(test_yml)
        test_loaded_yaml = loader.get_single_data()
    test_comment_blocks = CommentCollector(test_yml_file, test_loaded_yaml)
    test_comment_blocks.extract_all_comment_blocks()

    for ref_cmnt, test_cmnt in zip(ref_comment_blocks, test_comment_blocks):
        assert ref_cmnt == test_cmnt, 'Comment is not consistent.'

    os.remove(test_yml_file)


def test_yml2xml_comment_parsing():
    """To test comment that written in xml for element attributes, e.g.
    attribute 'rank' for 'dimension' element and attribute 'exists' for
    'NXentry' group element.
    """
    input_yml = 'tests/data/nyaml2nxdl/NXcomment_yaml2nxdl.yaml'
    ref_xml = 'tests/data/nyaml2nxdl/Ref_NXcomment_yaml2nxdl.nxdl.xml'
    test_xml = 'tests/data/nyaml2nxdl/NXcomment_yaml2nxdl.nxdl.xml'

    result = CliRunner().invoke(nyml2nxdl.launch_tool,
                                ['--input-file', input_yml])
    assert result.exit_code == 0

    ref_root = ET.parse(ref_xml).getroot()
    test_root = ET.parse(test_xml).getroot()

    def recursive_compare(ref_root, test_root):
        assert ref_root.attrib.items() == test_root.attrib.items(), ("Got different xml element"
                                                                     "Atribute.")
        if ref_root.text and test_root.text:
            assert ref_root.text.strip() == test_root.text.strip(), ("Got differen element text.")
        if len(ref_root) > 0 and len(test_root) > 0:
            for x, y in zip(ref_root, test_root):
                recursive_compare(x, y)

    recursive_compare(ref_root, test_root)

    os.remove(test_xml)



from ..globals import directories
from ..utils.copy import copydir


def impatient_args(parser):
    parser.add_argument(
        "--prepare",
        action="store_true",
        help="Create the build files for the NeXus Impatient Guide",
    )


def impatient_exec(args):
    if args.prepare:
        copydir(directories.impatient_source_root(), directories.impatient_build_root())




#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Developer comments."""

# pylint: disable=no-member


# ISSUE WITH HSPY DEFAULTS FOR MISSING VENDOR DOCS CAUSING NON-REPRODUCIBLE tSTAMPS
# already the example with Adrien Teutrie's *.emd file shows
# that the intended start_time and end_time as required by the appdef
# are practically not reliable because, tested for hyperspy==1.7.1,
# the times which are injected into the hyperspy metainfo object by the emd
# i/o routine are substantially relying on default rules. An example:
# First hyperspy asks FEI for an AcquisitionStartDatetime
# which is what intuitively one would consider as the relevant time
# however, if this metadata key is not included in the *.emd file,
# hyperspy just adds for "convenience" the local system time!
# however, this retrieval of the local system time depends heavily on
# which version of the underlying libraries and operation system is used !
# in a nutshell, what we ideally want is, to phrase it with python's
# datetime package "time-zone-aware" objects to interact with and convert
# times properly into one another
# https://www.bloomberg.com/company/stories/work-dates-time-python/
# but this is not what we face unfortunately, so consequently the parser
# should leave the field marked with a note that the relevant time
# was not reliably retrievable to avoid a potential proliferation of
# mistakes down the analysis pipeline

# this makes clear that to interpret and reproduce the content which
# an emd reader creates on a given emd file, the results will be different
# on different computers simply already because of these silent defaults
# so strictly speaking one would have to run the parser inside a
# container in the nomad oasis and not on the local host machine as if
# one would do usually with somebody having a local instance of an oasis
# and thus this local oasis would be used the local operating system and
# installation to retrieve the start time and end time data from

# from datetime import datetime
# import pytz
# timezone = pytz.timezone(dct["General"]['time_zone'])
# from dateutil import tz
# from dateutil.tz import *
# from datetime import *
# tz.tzlocal().tzname(datetime.today())
# see KabulTz example here https://docs.python.org/3/library/datetime.html


# ISSUES WITH ROUTING RAW and PROCESSED (META)DATA to SPECIFIC tSTAMPED HIERARCHIES
# ##MK::the key question is how parsable content
# (clss instances) from a hyperspy run stand in relation?
# for instance if we load a vendor file and hspy spits out
# it contains a HAADF, a sum EDS spectrum, and a EDS spectrum stack
# can we safely assume that these images have been taken
# in the same event? i.e. can we consider that the microscope
# was able enough when the HAADF was taken and thereafter
# the user acquired the spectrum/(a)? In general we cannot assume
# that this is the case. However, hyperspy implies that
# this is the case, given that hspy can in many cases simply
# also not build on or even parse timestamps and interpret them
# consequently, all we get is a specific view of the experiment and
# some key steps and metadata to its processing
# in summary, NXem is much more general than what is commonly
# available via interoperability improving tools like hyperspy

# These points substantiate that two key ingredients are missing
# a proper time(stamp) tracking of actions on the microscope and a
# agreed upon documented! format how to understand these time stamps
#   of course eventually anonymized wrt to who was the user
#   but at least a clear statement when these were taken
# a proper and much more detailed documentation of the
#   vendor file formats and their relation to the software library
#   which and when in which context the tool was running
#   so that hspy devlopers can extract
#   these time stamps and use them to route the data
#   into more useful interoperable respective places.



"""This test is dedicated generate_eln converter tool.
"""

# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


import os
from typing import Dict
from click import testing


import yaml
from pynxtools.eln_mapper import eln_mapper


def check_keys_from_two_dict(dict1: Dict, dict2: Dict):
    """Compare keys of two dicts.

    Parameters
    ----------
    dict1 : Dict
        Dict-1 to compare the key with Dict-2
    dict2 : Dict
        Dict-2 to compare the key with Dict-1
    """
    for (key1, val1), (key2, val2) in zip(dict1.items(), dict2.items()):
        assert key1 == key2, "Test and Ref yaml file have different keys."
        if isinstance(val1, dict) and isinstance(val2, dict):
            check_keys_from_two_dict(val1, val2)


def test_reader_eln(tmp_path):
    """Test eln that goes with reader.

    Parameters
    ----------
    tmp_path : pathlib.Path
        A temporary path that is created for pytest
    """

    local_dir = os.path.abspath(os.path.dirname(__file__))
    ref_file = os.path.join(local_dir, '../data/eln_mapper/eln.yaml')

    test_file = os.path.join(tmp_path, 'eln.yaml')
    cli_run = testing.CliRunner()
    cli_run.invoke(eln_mapper.get_eln, [
        "--nxdl",
        "NXmpes",
        "--skip-top-levels",
        1,
        "--output-file",
        test_file,
        "--eln-type",
        'eln'])

    with open(ref_file, encoding='utf-8', mode='r') as ref_f:
        ref_dict = yaml.safe_load(ref_f)

    with open(test_file, encoding='utf-8', mode='r') as test_f:
        test_dict = yaml.safe_load(test_f)

    check_keys_from_two_dict(ref_dict, test_dict)


def test_scheme_eln(tmp_path):
    """Test Eln that goes in Nomad

    Parameters
    ----------
    tmp_path : pathlib.Path
        A temporary path that is created for pytest
    """

    local_dir = os.path.abspath(os.path.dirname(__file__))
    ref_file = os.path.join(local_dir, '../data/eln_mapper/mpes.scheme.archive.yaml')

    test_file = os.path.join(tmp_path, '.scheme.archive.yaml')
    cli_run = testing.CliRunner()
    cli_run.invoke(eln_mapper.get_eln, [
        "--nxdl",
        "NXmpes",
        "--output-file",
        test_file,
        "--eln-type",
        'scheme_eln'])
    with open(ref_file, encoding='utf-8', mode='r') as ref_f:
        ref_dict = yaml.safe_load(ref_f)

    with open(test_file, encoding='utf-8', mode='r') as test_f:
        test_dict = yaml.safe_load(test_f)

    check_keys_from_two_dict(ref_dict, test_dict)



import pytest

from ..nxdl import find_definition
from ..nxdl import iter_definitions
from ..nxdl import nxdl_schema
from ..nxdl import validate_definition


def test_iter_definitions():
    all_files = set(iter_definitions())
    assert all_files
    base_files = set(iter_definitions("base_classes"))
    assert base_files
    assert not (base_files - all_files)


def test_find_definition():
    assert find_definition("NXroot")
    assert not find_definition("NXwrong")
    assert not find_definition("NXroot", "applications")


@pytest.fixture(scope="module")
def xml_schema():
    return nxdl_schema()


@pytest.mark.parametrize("nxdl_file", list(iter_definitions()))
def test_nxdl_syntax(nxdl_file, xml_schema):
    validate_definition(nxdl_file, xml_schema)
